from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import BaseModel


class MultigridWatcherSpec(BaseModel):
    source: Path
    label: str
    visit: str
    instrument_name: str
    skip_existing_processing: bool = False
    destination_overrides: Dict[Path, str] = {}
    rsync_restarts: List[str] = []
    visit_end_time: Optional[datetime] = None



from __future__ import annotations

import itertools
import time
from pathlib import Path

import mrcfile
import numpy as np
import xmltodict
import yaml
from rich.prompt import Confirm

TILT = 1
tilt_angle = itertools.cycle(range(-60, 70, 10))


def initialise(dummy_location: Path) -> Path:
    base = dummy_location / "murfey_dummy"
    base.mkdir()
    detector_dir = base / "Data"
    detector_dir.mkdir()
    (detector_dir / "Supervisor").mkdir()
    with open(base / "config.yaml", "w") as yaml_out:
        yaml.dump(
            {
                "m12": {
                    "acquisition_software": ["epu", "tomo"],
                    "data_directories": [str(detector_dir)],
                    "rsync_basepath": str(dummy_location),
                    "calibrations": {"dummy": 0},
                }
            },
            yaml_out,
        )
        print(
            f"To set Murfey configuration for server: export MURFEY_MACHINE_CONFIGURATION={base / 'config.yaml'}"
        )
    return base


def tomo_file_name() -> str:
    angle = next(tilt_angle)
    global TILT
    file_name = f"Position_{TILT}_[{angle}].mrc"
    if angle == 60:
        TILT += 1
    return file_name


def write_mrc(
    base_path: Path,
    session_dir: str,
    mrc_name: str,
    data_dir: str = "Data",
    size: tuple = (1024, 1024, 50),
):
    rand_data = np.random.randint(0, high=128, size=size, dtype=np.uint8)
    with mrcfile.new(base_path / data_dir / session_dir / mrc_name) as mrc:
        mrc.set_data(rand_data)


def write_xml(base_path: Path, session_dir: str, xml_name: str):
    xml_data = {
        "Acquisition": {
            "Info": {
                "ImageSize": {
                    "Width": 2048,
                    "Height": 2048,
                },
                "SensorPixelSize": {
                    "Width": 5e-11,
                    "Height": 5e-11,
                },
            }
        }
    }
    with open(base_path / "Data" / session_dir / xml_name, "w") as xml:
        xml.write(xmltodict.unparse(xml_data))


def generate_data(base_path: Path, timeout: int | None = None, pause: int = 10):
    if Confirm.ask("Begin simulated data acquisiton?"):
        start_time = time.time()
        while True:
            try:
                if timeout is not None:
                    if time.time() - start_time > timeout:
                        return
                tilt = TILT
                tfn = tomo_file_name()
                try:
                    write_mrc(
                        base_path,
                        "Supervisor",
                        f"Position_{tilt}.mrc",
                        data_dir="Data",
                        size=(128, 128),
                    )
                except ValueError:
                    pass
                write_mrc(base_path, "Supervisor", tfn)
                write_xml(base_path, "Supervisor", tfn.replace(".mrc", ".xml"))
                time.sleep(pause)
            except KeyboardInterrupt:
                return



from __future__ import annotations

import asyncio
import json
import logging
from datetime import datetime
from typing import Any, TypeVar, Union

from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from sqlmodel import Session, select

import murfey.server.prometheus as prom
from murfey.server.murfey_db import get_murfey_db_session
from murfey.util import sanitise
from murfey.util.db import ClientEnvironment

T = TypeVar("T")

ws = APIRouter(prefix="/ws", tags=["Websocket"])
log = logging.getLogger("murfey.server.websocket")


class ConnectionManager:
    def __init__(self):
        self.active_connections: dict[int | str, WebSocket] = {}

    async def connect(
        self,
        websocket: WebSocket,
        client_id: Union[int, str],
        register_client: bool = True,
    ):
        await websocket.accept()
        self.active_connections[client_id] = websocket
        if register_client:
            if not isinstance(client_id, int):
                raise ValueError(
                    "To register a client the client ID must be an integer"
                )
            self._register_new_client(client_id)

    @staticmethod
    def _register_new_client(client_id: int):
        log.debug(f"Registering new client with ID {client_id}")
        new_client = ClientEnvironment(client_id=client_id, connected=True)
        murfey_db: Session = next(get_murfey_db_session())
        murfey_db.add(new_client)
        murfey_db.commit()
        murfey_db.close()

    def disconnect(self, client_id: Union[int, str], unregister_client: bool = True):
        self.active_connections.pop(client_id)
        if unregister_client:
            murfey_db: Session = next(get_murfey_db_session())
            client_env = murfey_db.exec(
                select(ClientEnvironment).where(
                    ClientEnvironment.client_id == client_id
                )
            ).one()
            murfey_db.delete(client_env)
            murfey_db.commit()
            murfey_db.close()

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await self.active_connections[connection].send_text(message)


manager = ConnectionManager()


@ws.websocket("/test/{client_id}")
async def websocket_endpoint(websocket: WebSocket, client_id: int):
    await manager.connect(websocket, client_id)
    await manager.broadcast(f"Client {client_id} joined")
    try:
        while True:
            data = await websocket.receive_text()
            try:
                json_data: dict = json.loads(data)
                if json_data["type"] == "log":  # and isinstance(json_data, dict)
                    json_data.pop("type")
                    await forward_log(json_data, websocket)
            except Exception:
                await manager.broadcast(f"Client #{client_id} sent message {data}")
    except WebSocketDisconnect:
        log.info(f"Disconnecting Client {int(sanitise(str(client_id)))}")
        murfey_db = next(get_murfey_db_session())
        client_env = murfey_db.exec(
            select(ClientEnvironment).where(ClientEnvironment.client_id == client_id)
        ).one()
        prom.monitoring_switch.labels(visit=client_env.visit).set(0)
        manager.disconnect(client_id)
        await manager.broadcast(f"Client #{client_id} disconnected")


@ws.websocket("/connect/{client_id}")
async def websocket_connection_endpoint(
    websocket: WebSocket,
    client_id: Union[int, str],
):
    await manager.connect(websocket, client_id, register_client=False)
    await manager.broadcast(f"Client {client_id} joined")
    try:
        while True:
            data = await websocket.receive_text()
            try:
                json_data: dict = json.loads(data)
                if json_data.get("type") == "log":  # and isinstance(json_data, dict)
                    json_data.pop("type")
                    await forward_log(json_data, websocket)
                elif json_data.get("message") == "refresh":
                    await manager.broadcast(json.dumps(json_data))

            except Exception:
                await manager.broadcast(f"Client #{client_id} sent message {data}")
    except WebSocketDisconnect:
        log.info(f"Disconnecting Client {sanitise(str(client_id))}")
        manager.disconnect(client_id, unregister_client=False)
        await manager.broadcast(f"Client #{client_id} disconnected")


async def check_connections(active_connections: list[WebSocket]):
    log.info("Checking connections")
    for connection in active_connections:
        log.info("Checking response")
        try:
            await asyncio.wait_for(connection.receive(), timeout=10)
        except asyncio.TimeoutError:
            log.info(f"Disconnecting Client {connection[0]}")
            manager.disconnect(connection[0], connection[1])


async def forward_log(logrecord: dict[str, Any], websocket: WebSocket):
    record_name = logrecord["name"]
    logrecord.pop("msecs", None)
    logrecord.pop("relativeCreated", None)
    client_timestamp = logrecord.pop("created", 0)
    if client_timestamp:
        logrecord["client_time"] = datetime.fromtimestamp(client_timestamp).isoformat()
    logrecord["client_host"] = websocket.client.host
    logging.getLogger(record_name).handle(logging.makeLogRecord(logrecord))


@ws.delete("/test/{client_id}")
async def close_ws_connection(client_id: int):
    murfey_db: Session = next(get_murfey_db_session())
    client_env = murfey_db.exec(
        select(ClientEnvironment).where(ClientEnvironment.client_id == client_id)
    ).one()
    client_env.connected = False
    visit_name = client_env.visit
    murfey_db.add(client_env)
    murfey_db.commit()
    murfey_db.close()
    client_id_str = str(client_id).replace("\r\n", "").replace("\n", "")
    log.info(f"Disconnecting {client_id_str}")
    manager.disconnect(client_id)
    prom.monitoring_switch.labels(visit=visit_name).set(0)
    await manager.broadcast(f"Client #{client_id} disconnected")


@ws.delete("/connect/{client_id}")
async def close_unrecorded_ws_connection(client_id: Union[int, str]):
    client_id_str = str(client_id).replace("\r\n", "").replace("\n", "")
    log.info(f"Disconnecting {client_id_str}")
    manager.disconnect(client_id)
    await manager.broadcast(f"Client #{client_id} disconnected")



import logging
from pathlib import Path
from typing import Dict, NamedTuple, Optional, Tuple, Union

import xmltodict

logger = logging.getLogger("murfey.util.spa_metadata")


class FoilHoleInfo(NamedTuple):
    id: int
    grid_square_id: int
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""
    diameter: Optional[float] = None


class GridSquareInfo(NamedTuple):
    id: int
    x_location: Optional[float] = None
    y_location: Optional[float] = None
    x_stage_position: Optional[float] = None
    y_stage_position: Optional[float] = None
    readout_area_x: Optional[int] = None
    readout_area_y: Optional[int] = None
    thumbnail_size_x: Optional[int] = None
    thumbnail_size_y: Optional[int] = None
    pixel_size: Optional[float] = None
    image: str = ""
    tag: str = ""


def grid_square_from_file(f: Path) -> int:
    for p in f.parts:
        if p.startswith("GridSquare"):
            return int(p.split("_")[1])
    raise ValueError(f"Grid square ID could not be determined from path {f}")


def foil_hole_from_file(f: Path) -> int:
    return int(f.name.split("_")[1])


def get_grid_square_atlas_positions(xml_path: Path, grid_square: str = "") -> Dict[
    str,
    Tuple[
        Optional[int],
        Optional[int],
        Optional[float],
        Optional[float],
        Optional[int],
        Optional[int],
        Optional[float],
    ],
]:
    with open(
        xml_path,
        "r",
    ) as dm:
        atlas_data = xmltodict.parse(dm.read())
    tile_info = atlas_data["AtlasSessionXml"]["Atlas"]["TilesEfficient"]["_items"][
        "TileXml"
    ]
    gs_pix_positions: Dict[
        str,
        Tuple[
            Optional[int],
            Optional[int],
            Optional[float],
            Optional[float],
            Optional[int],
            Optional[int],
            Optional[float],
        ],
    ] = {}
    for ti in tile_info:
        try:
            nodes = ti["Nodes"]["KeyValuePairs"]
        except KeyError:
            continue
        required_key = ""
        for key in nodes.keys():
            if key.startswith("KeyValuePairOfintNodeXml"):
                required_key = key
                break
        if not required_key:
            continue
        for gs in nodes[required_key]:
            if not isinstance(gs, dict):
                continue
            if not grid_square or gs["key"] == grid_square:
                gs_pix_positions[gs["key"]] = (
                    int(float(gs["value"]["b:PositionOnTheAtlas"]["c:Center"]["d:x"])),
                    int(float(gs["value"]["b:PositionOnTheAtlas"]["c:Center"]["d:y"])),
                    float(gs["value"]["b:PositionOnTheAtlas"]["c:Physical"]["d:x"])
                    * 1e9,
                    float(gs["value"]["b:PositionOnTheAtlas"]["c:Physical"]["d:y"])
                    * 1e9,
                    int(
                        float(gs["value"]["b:PositionOnTheAtlas"]["c:Size"]["d:width"])
                    ),
                    int(
                        float(gs["value"]["b:PositionOnTheAtlas"]["c:Size"]["d:height"])
                    ),
                    float(gs["value"]["b:PositionOnTheAtlas"]["c:Rotation"]),
                )
                if grid_square:
                    break
    return gs_pix_positions


def grid_square_data(xml_path: Path, grid_square: int) -> GridSquareInfo:
    image_paths = list(
        (xml_path.parent.parent).glob(
            f"Images-Disc*/GridSquare_{grid_square}/GridSquare_*.jpg"
        )
    )
    logger.info(
        f"{len(image_paths)} images found when searching {str(xml_path.parent.parent)}"
    )
    if image_paths:
        image_paths.sort(key=lambda x: x.stat().st_ctime)
        image_path = image_paths[-1]
        with open(Path(image_path).with_suffix(".xml")) as gs_xml:
            gs_xml_data = xmltodict.parse(gs_xml.read())
        readout_area = gs_xml_data["MicroscopeImage"]["microscopeData"]["acquisition"][
            "camera"
        ]["ReadoutArea"]
        pixel_size = gs_xml_data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
            "numericValue"
        ]
        full_size = (int(readout_area["a:width"]), int(readout_area["a:height"]))
        return GridSquareInfo(
            id=grid_square,
            readout_area_x=full_size[0] if image_path else None,
            readout_area_y=full_size[1] if image_path else None,
            thumbnail_size_x=int((512 / max(full_size)) * full_size[0]),
            thumbnail_size_y=int((512 / max(full_size)) * full_size[1]),
            pixel_size=float(pixel_size) if image_path else None,
            image=str(image_path),
        )
    return GridSquareInfo(id=grid_square)


def foil_hole_data(xml_path: Path, foil_hole: int, grid_square: int) -> FoilHoleInfo:
    required_key = ""
    if xml_path.is_file():
        with open(xml_path, "r") as xml:
            for_parsing = xml.read()
            data = xmltodict.parse(for_parsing)
        data = data["GridSquareXml"]
        if "TargetLocationsEfficient" in data["TargetLocations"].keys():
            # Grids with regular foil holes
            serialization_array = data["TargetLocations"]["TargetLocationsEfficient"][
                "a:m_serializationArray"
            ]
        elif "TargetLocations" in data["TargetLocations"].keys():
            # Lacey grids
            serialization_array = data["TargetLocations"]["TargetLocations"][
                "a:m_serializationArray"
            ]
        else:
            logger.warning(f"Target locations not found for {str(xml_path)}")
            return FoilHoleInfo(id=foil_hole, grid_square_id=grid_square)

        if len(serialization_array.keys()) == 0:
            return FoilHoleInfo(id=foil_hole, grid_square_id=grid_square)
        for key in serialization_array.keys():
            if key.startswith("b:KeyValuePairOfintTargetLocation"):
                required_key = key
                break
    image_paths = list(
        (xml_path.parent.parent).glob(
            f"Images-Disc*/GridSquare_{grid_square}/FoilHoles/FoilHole_{foil_hole}_*.jpg"
        )
    )
    image_paths.sort(key=lambda x: x.stat().st_ctime)
    image_path: Union[Path, str] = image_paths[-1] if image_paths else ""
    if image_path:
        with open(Path(image_path).with_suffix(".xml")) as fh_xml:
            fh_xml_data = xmltodict.parse(fh_xml.read())
        readout_area = fh_xml_data["MicroscopeImage"]["microscopeData"]["acquisition"][
            "camera"
        ]["ReadoutArea"]
        pixel_size = fh_xml_data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
            "numericValue"
        ]
        full_size = (int(readout_area["a:width"]), int(readout_area["a:height"]))
    if required_key:
        for fh_block in serialization_array[required_key]:
            pix = fh_block["b:value"]["PixelCenter"]
            stage = fh_block["b:value"]["StagePosition"]
            diameter = fh_block["b:value"]["PixelWidthHeight"]["c:width"]
            if int(fh_block["b:key"]) == foil_hole:
                return FoilHoleInfo(
                    id=foil_hole,
                    grid_square_id=grid_square,
                    x_location=float(pix["c:x"]),
                    y_location=float(pix["c:y"]),
                    x_stage_position=float(stage["c:X"]),
                    y_stage_position=float(stage["c:Y"]),
                    readout_area_x=full_size[0] if image_path else None,
                    readout_area_y=full_size[1] if image_path else None,
                    thumbnail_size_x=None,
                    thumbnail_size_y=None,
                    pixel_size=float(pixel_size) if image_path else None,
                    image=str(image_path),
                    diameter=diameter,
                )
    elif image_path:
        return FoilHoleInfo(
            id=foil_hole,
            grid_square_id=grid_square,
            readout_area_x=full_size[0] if image_path else None,
            readout_area_y=full_size[1] if image_path else None,
            thumbnail_size_x=None,
            thumbnail_size_y=None,
            pixel_size=float(pixel_size) if image_path else None,
            image=str(image_path),
        )
    logger.warning(
        f"Foil hole positions could not be determined from metadata file {xml_path} for foil hole {foil_hole}"
    )
    return FoilHoleInfo(id=foil_hole, grid_square_id=grid_square)



import sys
from typing import Optional
from urllib.parse import urlparse

import pytest
import uvicorn
from fastapi import FastAPI
from fastapi.testclient import TestClient
from packaging.version import Version
from pytest_mock import MockerFixture

import murfey
from murfey.client.update import UPDATE_SUCCESS
from murfey.instrument_server import check_for_updates, start_instrument_server
from murfey.server.api.bootstrap import pypi as pypi_router
from murfey.server.api.bootstrap import version as version_router
from murfey.util.api import url_path_for

# Set up a test router with only the essential endpoints
app = FastAPI()
for router in [pypi_router, version_router]:
    app.include_router(router)
client = TestClient(app)
base_url = str(client.base_url)


check_for_updates_test_matrix = (
    # Downgrade, upgrade, or keep client version?
    ("downgrade",),
    ("upgrade",),
    ("keep",),
)


@pytest.mark.parametrize("test_params", check_for_updates_test_matrix)
def test_check_for_updates(
    test_params: tuple[str],
    mocker: MockerFixture,
):

    # Unpack test params
    (bump_client_version,) = test_params

    # Modify client version as needed
    current_version = murfey.__version__
    supported_client_version = murfey.__supported_client_version__

    major, minor, patch = Version(current_version).release

    # Adjust the perceived client version in the function being tested
    if bump_client_version == "downgrade":
        support_client_version_parts = Version(supported_client_version).release
        if patch == 0:
            if minor == 0:
                if major == 0:
                    # This can't be downgraded, so skip
                    pytest.skip("This version can't be downgraded anymore; skipping")
                else:
                    major = support_client_version_parts[0] - 1
                    print(f"Downgraded major version to {major}")
            else:
                minor = support_client_version_parts[1] - 1
                print(f"Downgraded minor version to {minor}")
        else:
            patch = support_client_version_parts[2] - 1
            print(f"Downgraded patch version to {patch}")
    elif bump_client_version == "upgrade":
        patch += 1
        print(f"Bumped patch version to {patch}")
    mock_client_version = f"{major}.{minor}.{patch}"

    # Run the version check query and get a response to patch in later
    api_base = urlparse(base_url, allow_fragments=False)
    proxy_path = api_base.path.rstrip("/")
    version_check_path = url_path_for("bootstrap.version", "get_version")
    version_check_query = f"client_version={mock_client_version}"
    version_check_url = api_base._replace(
        path=f"{proxy_path}{version_check_path}",
        query=version_check_query,
    )
    version_check_response = client.get(f"{version_check_path}?{version_check_query}")

    # Check that the endpoint works as expected
    assert version_check_response.status_code == 200
    assert version_check_response.json() == {
        "server": current_version,
        "oldest-supported-client": supported_client_version,
        "client-needs-update": True if bump_client_version == "downgrade" else False,
        "client-needs-downgrade": True if bump_client_version == "upgrade" else False,
    }

    # Patch the URL parse result
    mock_parse = mocker.patch("murfey.instrument_server.urlparse")
    mock_parse.return_value = api_base

    # Patch the result of get
    mock_get = mocker.patch("murfey.client.update.requests.get")
    mock_get.return_value = version_check_response

    # Patch the installation function
    mock_install = mocker.patch("murfey.client.update.install_murfey")

    # Patch the perceived client version
    mocker.patch("murfey.client.update.murfey.__version__", new=mock_client_version)

    # If changing the client version, check that 'install_murfey' and 'exit' are called
    if bump_client_version in ("upgrade", "downgrade"):
        with pytest.raises(SystemExit) as exc_info:
            check_for_updates()
        mock_install.assert_called_once()
        # Check that 'exit' is called with the correct message
        assert exc_info.value.code == UPDATE_SUCCESS
    # If client version is the same, 'install_murfey' shouldn't be called
    else:
        check_for_updates()
        mock_install.assert_not_called()

    # Check that the query URL is correct
    mock_get.assert_called_once_with(version_check_url.geturl())


start_instrument_server_test_matrix = (
    # Host | Port
    (
        None,
        None,
    ),  # Test default values
    (
        "127.0.0.1",
        8000,
    ),  # Test manually included values
)


@pytest.mark.parametrize("test_params", start_instrument_server_test_matrix)
def test_start_instrument_server(
    mocker: MockerFixture, test_params: tuple[Optional[str], Optional[int]]
):

    # Unpack test params
    host, port = test_params

    # Patch the Uvicorn Server instance
    mock_server = mocker.patch("uvicorn.Server")
    # Disable 'run'; we just want to confirm it's called correctly
    mock_server.run.return_value = lambda: None

    # Patch the websocket instance
    mock_wsapp = mocker.patch("murfey.client.websocket.WSApp")
    mock_wsapp.return_value = mocker.Mock()  # Disable functionality

    # Construct the expected Uvicorn Config object and save it as a dict
    expected_config = vars(
        uvicorn.Config(
            "murfey.instrument_server.main:app",
            host=host if host is not None else "0.0.0.0",
            port=port if port is not None else 8001,
            log_config=None,
            ws_ping_interval=300,
            ws_ping_timeout=300,
        )
    )

    # Construct the arguments to pass to the instrument server
    sys.argv = [
        "murfey.instrument_server",
    ]

    # Add host and port if they're present
    if host is not None:
        sys.argv.extend(["--host", host])
    if port is not None:
        sys.argv.extend(["--port", str(port)])

    # Run the function
    start_instrument_server()

    # Check that the server was called with the correct arguments
    args, kwargs = mock_server.call_args
    actual_config = vars(kwargs["config"])
    assert expected_config == actual_config



from __future__ import annotations

import logging
from pathlib import Path
from threading import RLock
from typing import Callable, Dict, List, OrderedDict

import requests
import xmltodict

import murfey.util.eer
from murfey.client.context import Context, ProcessingParameter
from murfey.client.instance_environment import (
    MovieID,
    MovieTracker,
    MurfeyID,
    MurfeyInstanceEnvironment,
)
from murfey.util.api import url_path_for
from murfey.util.client import (
    authorised_requests,
    capture_post,
    get_machine_config_client,
)
from murfey.util.mdoc import get_block, get_global_data, get_num_blocks

logger = logging.getLogger("murfey.client.contexts.tomo")

requests.get, requests.post, requests.put, requests.delete = authorised_requests()


def _get_tilt_angle_v5_7(p: Path) -> str:
    return p.name.split("[")[1].split("]")[0]


def _get_tilt_angle_v5_11(p: Path) -> str:
    _split = p.name.split("_")[2].split(".")
    return ".".join(_split[:-1])


def _find_angle_index(split_name: List[str]) -> int:
    for i, part in enumerate(split_name):
        if "." in part and part[-1].isnumeric():
            return i
    return -1


def _get_tilt_angle_v5_12(p: Path) -> str:
    split_name = p.stem.split("_")
    angle_idx = _find_angle_index(split_name)
    if angle_idx == -1:
        return ""
    return split_name[angle_idx]


tomo_tilt_info = {
    "5.7": _get_tilt_angle_v5_7,
    "5.11": _get_tilt_angle_v5_11,
    "5.12": _get_tilt_angle_v5_12,
}


def _construct_tilt_series_name(file_path: Path) -> str:
    # Assuming files end with _{tiltnumber}_{angle}_{date}_{time}_{fractions}.{suffix}
    split_name = file_path.name.split("_")
    return "_".join(split_name[:-5])


class TomographyContext(Context):
    user_params = [
        ProcessingParameter(
            "dose_per_frame", "Dose Per Frame (e- / Angstrom^2 / frame)", default=1
        ),
        ProcessingParameter("gain_ref", "Gain Reference"),
        ProcessingParameter("eer_fractionation", "EER Fractionation", default=20),
    ]
    metadata_params = [
        ProcessingParameter("voltage", "Voltage"),
        ProcessingParameter("image_size_x", "Image Size X"),
        ProcessingParameter("image_size_y", "Image Size Y"),
        ProcessingParameter("pixel_size_on_image", "Pixel Size"),
        ProcessingParameter("motion_corr_binning", "Motion Correction Binning"),
        ProcessingParameter("frame_count", "Number of image frames"),
        ProcessingParameter("tilt_axis", "Stage rotation angle"),
        ProcessingParameter("num_eer_frames", "Number of EER Frames"),
    ]

    def __init__(self, acquisition_software: str, basepath: Path):
        super().__init__("Tomography", acquisition_software)
        self._basepath = basepath
        self._tilt_series: Dict[str, List[Path]] = {}
        self._tilt_series_with_pjids: List[str] = []
        self._tilt_series_sizes: Dict[str, int] = {}
        self._completed_tilt_series: List[str] = []
        self._aligned_tilt_series: List[str] = []
        self._data_collection_stash: list = []
        self._processing_job_stash: dict = {}
        self._lock: RLock = RLock()

    def register_tomography_data_collections(
        self,
        file_extension: str,
        image_directory: str,
        environment: MurfeyInstanceEnvironment | None = None,
    ):
        if not environment:
            logger.error(
                "No environment passed to register tomography data collections"
            )
            return
        try:
            dcg_url = f"{str(environment.url.geturl())}{url_path_for('workflow.router', 'register_dc_group', visit_name=environment.visit, session_id=environment.murfey_session)}"
            dcg_data = {
                "experiment_type": "tomo",
                "experiment_type_id": 36,
                "tag": str(self._basepath),
                "atlas": "",
                "sample": None,
            }
            capture_post(dcg_url, json=dcg_data)

            for tilt_series in self._tilt_series.keys():
                if tilt_series not in self._tilt_series_with_pjids:
                    dc_url = f"{str(environment.url.geturl())}{url_path_for('workflow.router', 'start_dc', visit_name=environment.visit, session_id=environment.murfey_session)}"
                    dc_data = {
                        "experiment_type": "tomography",
                        "file_extension": file_extension,
                        "acquisition_software": self._acquisition_software,
                        "image_directory": image_directory,
                        "data_collection_tag": tilt_series,
                        "source": str(self._basepath),
                        "tag": tilt_series,
                    }
                    if (
                        self.data_collection_parameters
                        and self.data_collection_parameters.get("voltage")
                    ):
                        # Once mdoc parameters are known register processing jobs
                        dc_data.update(
                            {
                                "voltage": self.data_collection_parameters["voltage"],
                                "pixel_size_on_image": self.data_collection_parameters[
                                    "pixel_size_on_image"
                                ],
                                "image_size_x": self.data_collection_parameters[
                                    "image_size_x"
                                ],
                                "image_size_y": self.data_collection_parameters[
                                    "image_size_y"
                                ],
                                "magnification": self.data_collection_parameters[
                                    "magnification"
                                ],
                            }
                        )
                        capture_post(dc_url, json=dc_data)

                        proc_url = f"{str(environment.url.geturl())}{url_path_for('workflow.router', 'register_proc', visit_name=environment.visit, session_id=environment.murfey_session)}"
                        for recipe in ("em-tomo-preprocess", "em-tomo-align"):
                            capture_post(
                                proc_url,
                                json={
                                    "tag": tilt_series,
                                    "source": str(self._basepath),
                                    "recipe": recipe,
                                    "experiment_type": "tomography",
                                },
                            )
                        self._tilt_series_with_pjids.append(tilt_series)
                    else:
                        logger.info(
                            "Cannot register data collection yet as no values from mdoc"
                        )

        except Exception as e:
            logger.error(f"ERROR {e}, {self.data_collection_parameters}", exc_info=True)

    def _file_transferred_to(
        self, environment: MurfeyInstanceEnvironment, source: Path, file_path: Path
    ):
        machine_config = get_machine_config_client(
            str(environment.url.geturl()),
            instrument_name=environment.instrument_name,
            demo=environment.demo,
        )
        if environment.visit in environment.default_destinations[source]:
            return (
                Path(machine_config.get("rsync_basepath", ""))
                / Path(environment.default_destinations[source])
                / file_path.name
            )
        return (
            Path(machine_config.get("rsync_basepath", ""))
            / Path(environment.default_destinations[source])
            / environment.visit
            / file_path.name
        )

    def _get_source(
        self, file_path: Path, environment: MurfeyInstanceEnvironment
    ) -> Path | None:
        for s in environment.sources:
            if file_path.is_relative_to(s):
                return s
        return None

    def _add_tilt(
        self,
        file_path: Path,
        extract_tilt_angle: Callable[[Path], str],
        environment: MurfeyInstanceEnvironment | None = None,
        required_strings: List[str] | None = None,
    ) -> List[str]:
        if not environment:
            logger.warning("No environment passed in")
            return []
        source = self._get_source(file_path, environment)
        if not source:
            logger.warning(f"No source found for file {file_path}")
            return []
        required_strings = (
            ["fractions"] if required_strings is None else required_strings
        )
        if required_strings and not any(r in file_path.name for r in required_strings):
            return []
        try:
            tilt_angle = extract_tilt_angle(file_path)
            try:
                float(tilt_angle)
            except ValueError:
                logger.error(f"whoops, {tilt_angle}")
                return []
            tilt_series = _construct_tilt_series_name(file_path)

        except Exception:
            logger.debug(
                f"Tilt series and angle could not be determined for {file_path}"
            )
            return []

        if environment:
            file_transferred_to = self._file_transferred_to(
                environment, source, file_path
            )
            environment.movies[file_transferred_to] = MovieTracker(
                movie_number=next(MovieID),
                motion_correction_uuid=next(MurfeyID),
            )
            environment.movie_tilt_pair[file_transferred_to] = tilt_series
            if environment.tilt_angles.get(tilt_series):
                environment.tilt_angles[tilt_series].append(
                    [str(file_transferred_to), tilt_angle]
                )
            else:
                environment.tilt_angles[tilt_series] = [
                    [str(file_transferred_to), tilt_angle]
                ]
        if tilt_series in self._completed_tilt_series:
            logger.info(
                f"Tilt series {tilt_series} was previously thought complete but now {file_path} has been seen"
            )
            self._completed_tilt_series.remove(tilt_series)
            rerun_url = f"{str(environment.url.geturl())}{url_path_for('workflow.tomo_router', 'register_tilt_series_for_rerun', visit_name=environment.visit)}"
            rerun_data = {
                "session_id": environment.murfey_session,
                "tag": tilt_series,
                "source": str(file_path.parent),
            }
            capture_post(rerun_url, json=rerun_data)
            if tilt_series in self._aligned_tilt_series:
                with self._lock:
                    self._aligned_tilt_series.remove(tilt_series)

        if not self._tilt_series.get(tilt_series):
            logger.info(f"New tilt series found: {tilt_series}")
            self._tilt_series[tilt_series] = [file_path]
            ts_url = f"{str(environment.url.geturl())}{url_path_for('workflow.tomo_router', 'register_tilt_series', visit_name=environment.visit)}"
            ts_data = {
                "session_id": environment.murfey_session,
                "tag": tilt_series,
                "source": str(file_path.parent),
            }
            capture_post(ts_url, json=ts_data)
            if not self._tilt_series_sizes.get(tilt_series):
                self._tilt_series_sizes[tilt_series] = 0

            # Will register processing jobs for all tilt series except the first one
            self.register_tomography_data_collections(
                file_extension=file_path.suffix,
                image_directory=str(
                    environment.default_destinations.get(
                        file_path.parent, file_path.parent
                    )
                ),
                environment=environment,
            )
        else:
            if file_path not in self._tilt_series[tilt_series]:
                for p in self._tilt_series[tilt_series]:
                    if tilt_angle == extract_tilt_angle(p):
                        break
                else:
                    self._tilt_series[tilt_series].append(file_path)

        if environment:
            tilt_url = f"{str(environment.url.geturl())}{url_path_for('workflow.tomo_router', 'register_tilt', visit_name=environment.visit, session_id=environment.murfey_session)}"
            tilt_data = {
                "movie_path": str(file_transferred_to),
                "tilt_series_tag": tilt_series,
                "source": str(file_path.parent),
            }
            capture_post(tilt_url, json=tilt_data)

            eer_fractionation_file = None
            if self.data_collection_parameters.get("num_eer_frames"):
                response = requests.post(
                    f"{str(environment.url.geturl())}{url_path_for('file_io_instrument.router', 'write_eer_fractionation_file', visit_name=environment.visit, session_id=environment.murfey_session)}",
                    json={
                        "num_frames": self.data_collection_parameters["num_eer_frames"],
                        "fractionation": self.data_collection_parameters[
                            "eer_fractionation"
                        ],
                        "dose_per_frame": environment.dose_per_frame or 0,
                        "fractionation_file_name": "eer_fractionation_tomo.txt",
                    },
                )
                eer_fractionation_file = response.json()["eer_fractionation_file"]
            preproc_url = f"{str(environment.url.geturl())}{url_path_for('workflow.tomo_router', 'request_tomography_preprocessing', visit_name=environment.visit, session_id=environment.murfey_session)}"
            preproc_data = {
                "path": str(file_transferred_to),
                "description": "",
                "image_number": environment.movies[file_transferred_to].movie_number,
                "pixel_size": self.data_collection_parameters.get(
                    "pixel_size_on_image", 0
                ),
                "dose_per_frame": environment.dose_per_frame or 0,
                "frame_count": self.data_collection_parameters.get("frame_count", 0),
                "tilt_axis": self.data_collection_parameters.get("tilt_axis", 85),
                "mc_binning": self.data_collection_parameters.get(
                    "motion_corr_binning", 1
                ),
                "gain_ref": environment.gain_ref,
                "voltage": self.data_collection_parameters.get("voltage", 300),
                "eer_fractionation_file": eer_fractionation_file,
                "tag": tilt_series,
                "group_tag": str(self._basepath),
            }
            capture_post(preproc_url, json=preproc_data)

        return self._check_tilt_series(tilt_series)

    def _check_tilt_series(
        self,
        tilt_series: str,
    ) -> List[str]:
        newly_completed_series: List[str] = []
        mdoc_tilt_series_size = self._tilt_series_sizes.get(tilt_series, 0)
        if not self._tilt_series or not mdoc_tilt_series_size:
            logger.debug(f"Tilt series size not yet set for {tilt_series!r}")
            return newly_completed_series

        counted_tilts = len(self._tilt_series.get(tilt_series, []))
        tilt_series_size_check = counted_tilts >= mdoc_tilt_series_size
        if tilt_series_size_check and tilt_series not in self._completed_tilt_series:
            self._completed_tilt_series.append(tilt_series)
            newly_completed_series.append(tilt_series)
        else:
            logger.debug(
                f"{tilt_series!r} not complete yet. Counted {counted_tilts} tilts. "
                f"Expected number of tilts was {mdoc_tilt_series_size}"
            )
        return newly_completed_series

    def _add_tomo_tilt(
        self,
        file_path: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        required_strings: List[str] | None = None,
    ) -> List[str]:
        required_strings = (
            ["fractions"] if required_strings is None else required_strings
        )
        if not any(r in file_path.name for r in required_strings):
            return []
        if environment:
            if tomo_version := environment.software_versions.get("tomo"):
                tilt_info_extraction = tomo_tilt_info.get(tomo_version)
                if not tilt_info_extraction:
                    raise ValueError(
                        f"Extraction routines for TFS Tomo version {tomo_version} unknown"
                    )
            else:
                tilt_info_extraction = tomo_tilt_info["5.7"]
        else:
            tilt_info_extraction = tomo_tilt_info["5.7"]
        return self._add_tilt(
            file_path,
            tilt_info_extraction,
            environment=environment,
            required_strings=required_strings,
        )

    def post_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ) -> List[str]:
        super().post_transfer(
            transferred_file=transferred_file,
            environment=environment,
            **kwargs,
        )

        data_suffixes = (".mrc", ".tiff", ".tif", ".eer")
        completed_tilts = []

        if "gain" not in transferred_file.name:
            if transferred_file.suffix in data_suffixes:
                if self._acquisition_software == "tomo":
                    if environment:
                        machine_config = get_machine_config_client(
                            str(environment.url.geturl()),
                            instrument_name=environment.instrument_name,
                            demo=environment.demo,
                        )
                    else:
                        machine_config = {}
                    required_strings = (
                        machine_config.get("data_required_substrings", {})
                        .get("tomo", {})
                        .get(transferred_file.suffix, ["fractions"])
                    )
                    completed_tilts = self._add_tomo_tilt(
                        transferred_file,
                        environment=environment,
                        required_strings=kwargs.get("required_strings")
                        or required_strings,
                    )
                else:
                    logger.warning(f"Unknown data file {transferred_file}")
            if transferred_file.suffix == ".mdoc":
                with open(transferred_file, "r") as md:
                    tilt_series = transferred_file.stem
                    self._tilt_series_sizes[tilt_series] = get_num_blocks(md)
                if environment:
                    source = self._get_source(transferred_file, environment)
                    if source:
                        completed_tilts = self._check_tilt_series(tilt_series)

                    # Always update the tilt series length in the database after an mdoc
                    if environment.murfey_session is not None:
                        length_url = f"{str(environment.url.geturl())}{url_path_for('workflow.tomo_router', 'register_tilt_series_length', session_id=environment.murfey_session)}"
                        capture_post(
                            length_url,
                            json={
                                "tags": [tilt_series],
                                "source": str(transferred_file.parent),
                                "tilt_series_lengths": [
                                    self._tilt_series_sizes[tilt_series]
                                ],
                            },
                        )

        if completed_tilts and environment:
            logger.info(
                f"The following tilt series are considered complete: {completed_tilts} "
                f"after {transferred_file}"
            )
            complete_url = f"{str(environment.url.geturl())}{url_path_for('workflow.tomo_router', 'register_completed_tilt_series', visit_name=environment.visit, session_id=environment.murfey_session)}"
            capture_post(
                complete_url,
                json={
                    "tags": completed_tilts,
                    "source": str(transferred_file.parent),
                    "tilt_series_lengths": [
                        len(self._tilt_series.get(ts, [])) for ts in completed_tilts
                    ],
                },
            )
        return completed_tilts

    def post_first_transfer(
        self,
        transferred_file: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        **kwargs,
    ):
        self.post_transfer(transferred_file, environment=environment, **kwargs)

    def gather_metadata(
        self, metadata_file: Path, environment: MurfeyInstanceEnvironment | None = None
    ) -> OrderedDict:
        if metadata_file.suffix not in (".mdoc", ".xml"):
            raise ValueError(
                f"Tomography gather_metadata method expected xml or mdoc file not {metadata_file.name}"
            )
        try:
            if not metadata_file.is_file():
                logger.debug(f"Metadata file {metadata_file} not found")
                return OrderedDict({})
            if metadata_file.suffix == ".xml":
                with open(metadata_file, "r") as xml:
                    try:
                        for_parsing = xml.read()
                    except Exception:
                        logger.warning(
                            f"Failed to parse file {metadata_file}", exc_info=True
                        )
                        return OrderedDict({})
                    data = xmltodict.parse(for_parsing)
                try:
                    metadata: OrderedDict = OrderedDict({})
                    metadata["experiment_type"] = "tomography"
                    metadata["voltage"] = 300
                    metadata["image_size_x"] = data["Acquisition"]["Info"]["ImageSize"][
                        "Width"
                    ]
                    metadata["image_size_y"] = data["Acquisition"]["Info"]["ImageSize"][
                        "Height"
                    ]
                    metadata["pixel_size_on_image"] = float(
                        data["Acquisition"]["Info"]["SensorPixelSize"]["Height"]
                    )
                    metadata["motion_corr_binning"] = 1
                    metadata["gain_ref"] = None
                    metadata["dose_per_frame"] = (
                        environment.dose_per_frame if environment else None
                    )
                    metadata["source"] = str(self._basepath)
                except KeyError:
                    return OrderedDict({})
                return metadata
            with open(metadata_file, "r") as md:
                mdoc_data = get_global_data(md)
                num_blocks = get_num_blocks(md)
                md.seek(0)
                blocks = [get_block(md) for i in range(num_blocks)]
                mdoc_data_block = blocks[0]
            if not mdoc_data:
                return OrderedDict({})
            mdoc_metadata: OrderedDict = OrderedDict({})
            mdoc_metadata["experiment_type"] = "tomography"
            mdoc_metadata["voltage"] = float(mdoc_data["Voltage"])
            mdoc_metadata["frame_count"] = int(mdoc_data_block["NumSubFrames"])
            mdoc_metadata["tilt_axis"] = float(mdoc_data_block["RotationAngle"])
            mdoc_metadata["image_size_x"] = int(mdoc_data["ImageSize"][0])
            mdoc_metadata["image_size_y"] = int(mdoc_data["ImageSize"][1])
            mdoc_metadata["magnification"] = int(mdoc_data_block["Magnification"])
            superres_binning = int(mdoc_data_block["Binning"])
            binning_factor = 1
            if environment:
                server_config = requests.get(
                    f"{str(environment.url.geturl())}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=environment.instrument_name)}"
                ).json()
                if (
                    server_config.get("superres")
                    and superres_binning == 1
                    and environment.superres
                ):
                    binning_factor = 2
                ps_from_mag = (
                    server_config.get("calibrations", {})
                    .get("magnification", {})
                    .get(mdoc_data_block["Magnification"])
                )
                if ps_from_mag:
                    mdoc_metadata["pixel_size_on_image"] = (
                        float(ps_from_mag) * 1e-10 / binning_factor
                    )
            if mdoc_metadata.get("pixel_size_on_image") is None:
                mdoc_metadata["pixel_size_on_image"] = (
                    float(mdoc_data["PixelSpacing"]) * 1e-10
                )
            mdoc_metadata["motion_corr_binning"] = binning_factor
            mdoc_metadata["gain_ref"] = environment.gain_ref if environment else None
            mdoc_metadata["dose_per_frame"] = (
                environment.dose_per_frame if environment else None
            )
            mdoc_metadata["source"] = str(self._basepath)
            mdoc_metadata["tag"] = str(self._basepath)
            mdoc_metadata["tilt_series_tag"] = metadata_file.stem
            mdoc_metadata["exposure_time"] = float(mdoc_data_block["ExposureTime"])
            slit_width = mdoc_data_block["FilterSlitAndLoss"][0]
            if slit_width.lower() != "nan":
                mdoc_metadata["slit_width"] = float(slit_width)
            mdoc_metadata["file_extension"] = (
                f".{mdoc_data_block['SubFramePath'].split('.')[-1]}"
            )
            mdoc_metadata["eer_fractionation"] = (
                environment.eer_fractionation if environment else None
            ) or 20

            data_file = mdoc_data_block["SubFramePath"].split("\\")[-1]
            if data_file.split(".")[-1] == "eer":
                mdoc_metadata["num_eer_frames"] = murfey.util.eer.num_frames(
                    metadata_file.parent / data_file
                )
                mdoc_metadata["frame_count"] = int(
                    int(mdoc_metadata["eer_fractionation"])
                    / int(mdoc_metadata["num_eer_frames"])
                )
        except Exception as e:
            logger.error(
                f"Exception encountered in metadata gathering: {str(e)}", exc_info=True
            )
            return OrderedDict({})

        return mdoc_metadata



from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest

from murfey.server.ispyb import TransportManager
from murfey.util.config import MachineConfig
from murfey.workflows.clem.align_and_merge import submit_cluster_request

# Folder and file settings
session_id = 0
instrument_name = "clem"
raw_folder = "images"
processed_folder = "processed"
visit_name = "cm12345-6"
area_name = "test_area"
series_name = "test_series"
colors = [
    "gray",
    "green",
    "red",
]
feedback_queue = "murfey_feedback"

# Align and merge settings
crop_to_n_frames = 20
align_self = "enabled"
flatten = "max"
align_across = "enabled"


@pytest.fixture
def processed_dir(tmp_path: Path):
    processed_dir = tmp_path / visit_name / processed_folder
    processed_dir.mkdir(parents=True, exist_ok=True)
    return processed_dir


@pytest.fixture
def image_stacks(processed_dir: Path):

    image_dir = processed_dir / area_name / series_name
    image_dir.mkdir(parents=True, exist_ok=True)

    images = [image_dir / f"{color}.tiff" for color in colors]
    for image in images:
        if not image.exists():
            image.touch()

    return images


@pytest.fixture
def metadata(processed_dir: Path):

    metadata_dir = processed_dir / area_name / series_name / "metadata"
    metadata_dir.mkdir(parents=True, exist_ok=True)

    metadata = metadata_dir / f"{series_name}.xml"
    if not metadata.exists():
        metadata.touch()

    return metadata


@patch("murfey.workflows.clem.align_and_merge.get_machine_config")
def test_submit_cluster_request(
    mock_get_machine_config,
    image_stacks: list[Path],
    metadata: Path,
    processed_dir: Path,
):

    # Construct the long series name
    series_name_long = "--".join(
        image_stacks[0].parent.relative_to(processed_dir).parts
    )

    # Create a mock tranpsort object
    mock_transport = MagicMock(spec=TransportManager)
    mock_transport.feedback_queue = feedback_queue

    # Construct a mock MachineConfig object for use within the function
    mock_machine_config = MagicMock(spec=MachineConfig)
    mock_machine_config.processed_directory_name = processed_folder
    mock_get_machine_config.return_value = {
        instrument_name: mock_machine_config,
    }

    # Run the function
    submit_cluster_request(
        session_id=session_id,
        instrument_name=instrument_name,
        series_name=series_name_long,
        images=image_stacks,
        metadata=metadata,
        crop_to_n_frames=crop_to_n_frames,
        align_self=align_self,
        flatten=flatten,
        align_across=align_across,
        messenger=mock_transport,
    )

    # Construct expected recipe to be sent
    sent_recipe = {
        "recipes": ["clem-align-and-merge"],
        "parameters": {
            # Job parameters
            "series_name": series_name_long,
            "images": [str(file) for file in image_stacks],
            "metadata": str(metadata),
            "crop_to_n_frames": crop_to_n_frames,
            "align_self": align_self,
            "flatten": flatten,
            "align_across": align_across,
            # Other recipe parameters
            "session_dir": str(processed_dir.parent),
            "session_id": session_id,
            "job_name": series_name_long,
            "feedback_queue": feedback_queue,
        },
    }

    # Check that it sends the expected recipe
    mock_transport.send.assert_called_once_with(
        "processing_recipe",
        sent_recipe,
        new_connection=True,
    )



import logging
from pathlib import Path
from typing import Optional

import requests
import xmltodict

from murfey.client.context import Context
from murfey.client.contexts.spa import _file_transferred_to, _get_source
from murfey.client.contexts.spa_metadata import _atlas_destination
from murfey.client.instance_environment import MurfeyInstanceEnvironment, SampleInfo
from murfey.util.api import url_path_for
from murfey.util.client import authorised_requests, capture_post

logger = logging.getLogger("murfey.client.contexts.tomo_metadata")

requests.get, requests.post, requests.put, requests.delete = authorised_requests()


def ensure_dcg_exists(transferred_file: Path, environment: MurfeyInstanceEnvironment):
    # Make sure we have a data collection group
    source = _get_source(transferred_file, environment=environment)
    if not source:
        return None
    dcg_tag = str(source).replace(f"/{environment.visit}", "")
    url = f"{str(environment.url.geturl())}{url_path_for('workflow.router', 'register_dc_group', visit_name=environment.visit, session_id=environment.murfey_session)}"
    dcg_data = {
        "experiment_type": "tomo",
        "experiment_type_id": 36,
        "tag": dcg_tag,
    }
    capture_post(url, json=dcg_data)
    return dcg_tag


class TomographyMetadataContext(Context):
    def __init__(self, acquisition_software: str, basepath: Path):
        super().__init__("Tomography_metadata", acquisition_software)
        self._basepath = basepath

    def post_transfer(
        self,
        transferred_file: Path,
        environment: Optional[MurfeyInstanceEnvironment] = None,
        **kwargs,
    ):
        super().post_transfer(
            transferred_file=transferred_file,
            environment=environment,
            **kwargs,
        )

        if transferred_file.name == "Session.dm" and environment:
            logger.info("Tomography session metadata found")
            with open(transferred_file, "r") as session_xml:
                session_data = xmltodict.parse(session_xml.read())

            windows_path = session_data["TomographySession"]["AtlasId"]
            logger.info(f"Windows path to atlas metadata found: {windows_path}")
            if not windows_path:
                logger.warning("No atlas metadata path found")
                return
            visit_index = windows_path.split("\\").index(environment.visit)
            partial_path = "/".join(windows_path.split("\\")[visit_index + 1 :])
            logger.info("Partial Linux path successfully constructed from Windows path")

            source = _get_source(transferred_file, environment)
            if not source:
                logger.warning(
                    f"Source could not be identified for {str(transferred_file)}"
                )
                return

            source_visit_dir = source.parent

            logger.info(
                f"Looking for atlas XML file in metadata directory {str((source_visit_dir / partial_path).parent)}"
            )
            atlas_xml_path = list(
                (source_visit_dir / partial_path).parent.glob("Atlas_*.xml")
            )[0]
            logger.info(f"Atlas XML path {str(atlas_xml_path)} found")
            with open(atlas_xml_path, "rb") as atlas_xml:
                atlas_xml_data = xmltodict.parse(atlas_xml)
                atlas_pixel_size = float(
                    atlas_xml_data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
                        "numericValue"
                    ]
                )

            for p in partial_path.split("/"):
                if p.startswith("Sample"):
                    sample = int(p.replace("Sample", ""))
                    break
            else:
                logger.warning(f"Sample could not be identified for {transferred_file}")
                return
            environment.samples[source] = SampleInfo(
                atlas=Path(partial_path), sample=sample
            )
            url = f"{str(environment.url.geturl())}{url_path_for('workflow.router', 'register_dc_group', visit_name=environment.visit, session_id=environment.murfey_session)}"
            dcg_tag = "/".join(
                p for p in transferred_file.parent.parts if p != environment.visit
            ).replace("//", "/")
            dcg_data = {
                "experiment_type": "tomo",
                "experiment_type_id": 36,
                "tag": dcg_tag,
                "atlas": str(
                    _atlas_destination(environment, source, transferred_file)
                    / environment.samples[source].atlas.parent
                    / atlas_xml_path.with_suffix(".jpg").name
                ),
                "sample": environment.samples[source].sample,
                "atlas_pixel_size": atlas_pixel_size,
            }
            capture_post(url, json=dcg_data)

        elif transferred_file.name == "SearchMap.xml" and environment:
            logger.info("Tomography session search map xml found")
            dcg_tag = ensure_dcg_exists(transferred_file, environment)
            with open(transferred_file, "r") as sm_xml:
                sm_data = xmltodict.parse(sm_xml.read())

            # This bit gets SearchMap location on Atlas
            sm_pixel_size = float(
                sm_data["MicroscopeImage"]["SpatialScale"]["pixelSize"]["x"][
                    "numericValue"
                ]
            )
            stage_position = sm_data["MicroscopeImage"]["microscopeData"]["stage"][
                "Position"
            ]
            sm_binning = float(
                sm_data["MicroscopeImage"]["microscopeData"]["acquisition"]["camera"][
                    "Binning"
                ]["a:x"]
            )

            # Get the stage transformation
            sm_transformations = sm_data["MicroscopeImage"]["CustomData"][
                "a:KeyValueOfstringanyType"
            ]
            stage_matrix: dict[str, float] = {}
            image_matrix: dict[str, float] = {}
            for key_val in sm_transformations:
                if key_val["a:Key"] == "ReferenceCorrectionForStage":
                    stage_matrix = {
                        "m11": float(key_val["a:Value"]["b:_m11"]),
                        "m12": float(key_val["a:Value"]["b:_m12"]),
                        "m21": float(key_val["a:Value"]["b:_m21"]),
                        "m22": float(key_val["a:Value"]["b:_m22"]),
                    }
                elif key_val["a:Key"] == "ReferenceCorrectionForImageShift":
                    image_matrix = {
                        "m11": float(key_val["a:Value"]["b:_m11"]),
                        "m12": float(key_val["a:Value"]["b:_m12"]),
                        "m21": float(key_val["a:Value"]["b:_m21"]),
                        "m22": float(key_val["a:Value"]["b:_m22"]),
                    }
            if not stage_matrix or not image_matrix:
                logger.error(
                    f"No stage or image shift matrix found for {transferred_file}"
                )

            ref_matrix = {
                "m11": float(
                    sm_data["MicroscopeImage"]["ReferenceTransformation"]["matrix"][
                        "a:_m11"
                    ]
                ),
                "m12": float(
                    sm_data["MicroscopeImage"]["ReferenceTransformation"]["matrix"][
                        "a:_m12"
                    ]
                ),
                "m21": float(
                    sm_data["MicroscopeImage"]["ReferenceTransformation"]["matrix"][
                        "a:_m21"
                    ]
                ),
                "m22": float(
                    sm_data["MicroscopeImage"]["ReferenceTransformation"]["matrix"][
                        "a:_m22"
                    ]
                ),
            }

            source = _get_source(transferred_file, environment=environment)
            image_path = (
                _file_transferred_to(
                    environment, source, transferred_file.parent / "SearchMap.jpg"
                )
                if source
                else ""
            )

            sm_url = f"{str(environment.url.geturl())}{url_path_for('session_control.tomo_router', 'register_search_map', session_id=environment.murfey_session, sm_name=transferred_file.parent.name)}"
            capture_post(
                sm_url,
                json={
                    "tag": dcg_tag,
                    "x_stage_position": float(stage_position["X"]),
                    "y_stage_position": float(stage_position["Y"]),
                    "pixel_size": sm_pixel_size,
                    "image": str(image_path),
                    "binning": sm_binning,
                    "reference_matrix": ref_matrix,
                    "stage_correction": stage_matrix,
                    "image_shift_correction": image_matrix,
                },
            )

        elif transferred_file.name == "SearchMap.dm" and environment:
            logger.info("Tomography session search map dm found")
            dcg_tag = ensure_dcg_exists(transferred_file, environment)
            with open(transferred_file, "r") as sm_xml:
                sm_data = xmltodict.parse(sm_xml.read())

            # This bit gets SearchMap size
            try:
                sm_width = int(sm_data["TileSetXml"]["ImageSize"]["a:width"])
                sm_height = int(sm_data["TileSetXml"]["ImageSize"]["a:height"])
            except KeyError:
                logger.warning(f"Unable to find size for SearchMap {transferred_file}")
                readout_width = int(
                    sm_data["TileSetXml"]["AcquisitionSettings"]["a:camera"][
                        "a:ReadoutArea"
                    ]["b:width"]
                )
                readout_height = int(
                    sm_data["TileSetXml"]["AcquisitionSettings"]["a:camera"][
                        "a:ReadoutArea"
                    ]["b:height"]
                )
                sm_width = int(
                    8005 * readout_width / max(readout_height, readout_width)
                )
                sm_height = int(
                    8005 * readout_height / max(readout_height, readout_width)
                )
                logger.warning(
                    f"Inserting incorrect width {sm_width}, height {sm_height} for SearchMap display"
                )

            sm_url = f"{str(environment.url.geturl())}{url_path_for('session_control.tomo_router', 'register_search_map', session_id=environment.murfey_session, sm_name=transferred_file.parent.name)}"
            capture_post(
                sm_url,
                json={
                    "tag": dcg_tag,
                    "height": sm_height,
                    "width": sm_width,
                },
            )

        elif transferred_file.name == "BatchPositionsList.xml" and environment:
            logger.info("Tomography session batch positions list found")
            dcg_tag = ensure_dcg_exists(transferred_file, environment)
            with open(transferred_file) as xml:
                for_parsing = xml.read()
            batch_xml = xmltodict.parse(for_parsing)

            batch_positions_from_xml = batch_xml["BatchPositionsList"]["BatchPositions"]
            if not batch_positions_from_xml:
                logger.info("No batch positions yet")
                return

            batch_positions_list = batch_positions_from_xml["BatchPositionParameters"]
            if isinstance(batch_positions_list, dict):
                # Case of a single batch
                batch_positions_list = [batch_positions_list]

            for batch_position in batch_positions_list:
                batch_name = batch_position["Name"]
                search_map_name = batch_position["PositionOnTileSet"]["TileSetName"]
                batch_stage_location_x = float(
                    batch_position["PositionOnTileSet"]["StagePositionX"]
                )
                batch_stage_location_y = float(
                    batch_position["PositionOnTileSet"]["StagePositionY"]
                )

                # Always need search map before batch position
                sm_url = f"{str(environment.url.geturl())}{url_path_for('session_control.tomo_router', 'register_search_map', session_id=environment.murfey_session, sm_name=search_map_name)}"
                capture_post(
                    sm_url,
                    json={
                        "tag": dcg_tag,
                    },
                )

                # Then register batch position
                bp_url = f"{str(environment.url.geturl())}{url_path_for('session_control.tomo_router', 'register_batch_position', session_id=environment.murfey_session, batch_name=batch_name)}"
                capture_post(
                    bp_url,
                    json={
                        "tag": dcg_tag,
                        "x_stage_position": batch_stage_location_x,
                        "y_stage_position": batch_stage_location_y,
                        "x_beamshift": 0,
                        "y_beamshift": 0,
                        "search_map_name": search_map_name,
                    },
                )

                # Beamshifts
                if batch_position.get("AdditionalExposureTemplateAreas"):
                    beamshifts = batch_position["AdditionalExposureTemplateAreas"][
                        "ExposureTemplateAreaParameters"
                    ]
                    if type(beamshifts) is dict:
                        beamshifts = [beamshifts]
                    for beamshift in beamshifts:
                        beamshift_name = beamshift["Name"]
                        beamshift_position_x = float(beamshift["PositionX"])
                        beamshift_position_y = float(beamshift["PositionY"])

                        # Registration of beamshifted position
                        bp_url = f"{str(environment.url.geturl())}{url_path_for('session_control.tomo_router', 'register_batch_position', session_id=environment.murfey_session, batch_name=beamshift_name)}"
                        capture_post(
                            bp_url,
                            json={
                                "tag": dcg_tag,
                                "x_stage_position": batch_stage_location_x,
                                "y_stage_position": batch_stage_location_y,
                                "x_beamshift": beamshift_position_x,
                                "y_beamshift": beamshift_position_y,
                                "search_map_name": search_map_name,
                            },
                        )



from pathlib import Path
from typing import Optional
from unittest.mock import ANY, MagicMock, patch
from urllib.parse import urlparse

import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from murfey.instrument_server.api import _get_murfey_url
from murfey.instrument_server.api import router as client_router
from murfey.instrument_server.api import validate_session_token
from murfey.util import posix_path
from murfey.util.api import url_path_for


def set_up_test_client(session_id: Optional[int] = None):
    """
    Helper function to set up a test client for the instrument server with validation
    checks disabled.
    """
    # Set up the instrument server
    client_app = FastAPI()
    if session_id:
        client_app.dependency_overrides[validate_session_token] = lambda: session_id
    client_app.include_router(client_router)
    return TestClient(client_app)


test_get_murfey_url_params_matrix = (
    # Server URL to use
    ("default",),
    ("0.0.0.0:8000",),
    ("murfey_server",),
    ("http://murfey_server:8000",),
    ("http://murfey_server:8080/api",),
)


@pytest.mark.parametrize("test_params", test_get_murfey_url_params_matrix)
def test_get_murfey_url(
    test_params: tuple[str],
    mock_client_configuration,  # From conftest.py
):
    # Unpack test_params
    (server_url_to_test,) = test_params

    # Replace the server URL from the fixture with other ones for testing
    if server_url_to_test != "default":
        mock_client_configuration["Murfey"]["server"] = server_url_to_test

    # Mock the module-wide config variable with the fixture value
    # The fixture is only loaded within the test function, so this patch
    # has to happen inside the function instead of as a decorator
    with patch("murfey.instrument_server.api.config", mock_client_configuration):
        known_server = _get_murfey_url()

    # Prepend 'http://' to config URLs that don't have it for the comparison
    # Otherwise, urlparse stores it under the 'path' attribute
    original_url = str(mock_client_configuration["Murfey"].get("server"))
    if not original_url.startswith(("http://", "https://")):
        original_url = f"http://{original_url}"

    # Check that the components of the result match those in the config
    parsed_original = urlparse(original_url)
    parsed_server = urlparse(known_server)
    assert parsed_server.scheme in ("http", "https")
    assert parsed_server.hostname == parsed_original.hostname
    assert parsed_server.port == parsed_original.port
    assert parsed_server.netloc == parsed_original.netloc
    assert parsed_server.path == parsed_original.path


def test_check_multigrid_controller_status(mocker: MockerFixture):
    session_id = 1

    # Patch out the multigrid controllers that have been stored in memory
    mock_controller = MagicMock()
    mock_controller.dormant = False
    mock_controller.finalising = False
    mocker.patch(
        "murfey.instrument_server.api.controllers", {session_id: mock_controller}
    )

    # Set up the test client
    client_server = set_up_test_client(session_id=session_id)
    url_path = url_path_for(
        "api.router", "check_multigrid_controller_status", session_id=session_id
    )
    response = client_server.get(url_path)

    # Check that the result is as expected
    assert response.status_code == 200
    assert response.json() == {
        "dormant": False,
        "exists": True,
        "finalising": False,
    }


test_upload_gain_reference_params_matrix = (
    # Rsync URL settings
    ("http://1.1.1.1",),  # When rsync_url is provided
    ("",),  # When rsync_url is blank
    (None,),  # When rsync_url not provided
)


@pytest.mark.parametrize("test_params", test_upload_gain_reference_params_matrix)
def test_upload_gain_reference(
    mocker: MockerFixture,
    test_params: tuple[Optional[str]],
):
    # Unpack test parameters and define other ones
    (rsync_url_setting,) = test_params
    server_url = "https://murfey.server.test"
    instrument_name = "murfey"
    session_id = 1

    # Mock out objects
    mock_request = mocker.patch("murfey.instrument_server.api.requests")
    mock_get_server_url = mocker.patch("murfey.instrument_server.api._get_murfey_url")
    mock_subprocess = mocker.patch("murfey.instrument_server.api.subprocess")
    mocker.patch("murfey.instrument_server.api.tokens", {session_id: ANY})

    # Create a mock machine config base on the test params
    rsync_module = "data"
    gain_ref_dir = "C:/ProgramData/Gatan/Gain Reference"
    mock_machine_config = {
        "rsync_module": rsync_module,
        "gain_reference_directory": gain_ref_dir,
    }
    if rsync_url_setting is not None:
        mock_machine_config["rsync_url"] = rsync_url_setting

    # Assign expected values to the mock objects
    mock_response = MagicMock()
    mock_response.status_code = 200
    mock_response.json.return_value = mock_machine_config
    mock_request.get.return_value = mock_response
    mock_get_server_url.return_value = server_url
    mock_subprocess.run.return_value = MagicMock(returncode=0)

    # Construct payload and pass request to function
    gain_ref_file = f"{gain_ref_dir}/gain.mrc"
    visit_path = "2025/aa00000-0"
    gain_dest_dir = "processing"
    payload = {
        "gain_path": gain_ref_file,
        "visit_path": visit_path,
        "gain_destination_dir": gain_dest_dir,
    }

    # Set up instrument server test client
    client_server = set_up_test_client(session_id=session_id)

    # Poke the endpoint with the expected data
    url_path = url_path_for(
        "api.router",
        "upload_gain_reference",
        instrument_name=instrument_name,
        session_id=session_id,
    )
    response = client_server.post(url_path, json=payload)

    # Check that the machine config request was called
    machine_config_url = f"{server_url}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=instrument_name)}"
    mock_request.get.assert_called_once_with(
        machine_config_url,
        headers={"Authorization": ANY},
    )

    # Check that the subprocess was run with the expected arguments
    # If no rsync_url key is provided, or rsync_url key is empty,
    # It should default to the server URL
    expected_rsync_url = (
        urlparse(server_url) if not rsync_url_setting else urlparse(rsync_url_setting)
    )
    expected_rsync_path = f"{expected_rsync_url.hostname}::{rsync_module}/{visit_path}/{gain_dest_dir}/gain.mrc"
    expected_rsync_cmd = [
        "rsync",
        posix_path(Path(gain_ref_file)),
        expected_rsync_path,
    ]
    mock_subprocess.run.assert_called_once_with(
        expected_rsync_cmd,
        capture_output=True,
        text=True,
    )

    # Check that the function ran through to completion successfully
    assert response.json() == {"success": True}



from sqlmodel import Session

from murfey.server.api.session_control import count_number_of_movies
from murfey.util.db import (
    AutoProcProgram,
    DataCollection,
    DataCollectionGroup,
    Movie,
    MurfeyLedger,
    ProcessingJob,
)
from tests.conftest import ExampleVisit, get_or_create_db_entry


def test_movie_count(
    murfey_db_session: Session,  # From conftest.py
):

    # Insert table dependencies
    dcg_entry: DataCollectionGroup = get_or_create_db_entry(
        murfey_db_session,
        DataCollectionGroup,
        lookup_kwargs={
            "id": 0,
            "session_id": ExampleVisit.murfey_session_id,
            "tag": "test_dcg",
        },
    )
    dc_entry: DataCollection = get_or_create_db_entry(
        murfey_db_session,
        DataCollection,
        lookup_kwargs={
            "id": 0,
            "tag": "test_dc",
            "dcg_id": dcg_entry.id,
        },
    )
    processing_job_entry: ProcessingJob = get_or_create_db_entry(
        murfey_db_session,
        ProcessingJob,
        lookup_kwargs={
            "id": 0,
            "recipe": "test_recipe",
            "dc_id": dc_entry.id,
        },
    )
    autoproc_entry: AutoProcProgram = get_or_create_db_entry(
        murfey_db_session,
        AutoProcProgram,
        lookup_kwargs={
            "id": 0,
            "pj_id": processing_job_entry.id,
        },
    )

    # Insert test movies and one-to-one dependencies
    tag = "test_movie"
    num_movies = 5
    for i in range(num_movies):
        murfey_ledger_entry: MurfeyLedger = get_or_create_db_entry(
            murfey_db_session,
            MurfeyLedger,
            lookup_kwargs={
                "id": i,
                "app_id": autoproc_entry.id,
            },
        )
        _: Movie = get_or_create_db_entry(
            murfey_db_session,
            Movie,
            lookup_kwargs={
                "murfey_id": murfey_ledger_entry.id,
                "path": "/some/path",
                "image_number": i,
                "tag": tag,
            },
        )

    # Run function and evaluate result
    result = count_number_of_movies(murfey_db_session)
    assert result == {tag: num_movies}



from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import Dict, List, Optional

from fastapi import APIRouter, Depends, Request
from fastapi.responses import FileResponse
from pydantic import BaseModel
from sqlmodel import select
from werkzeug.utils import secure_filename

import murfey.server.api.websocket as ws
from murfey.server import _transport_object
from murfey.server.api import templates
from murfey.server.api.auth import MurfeyInstrumentNameFrontend as MurfeyInstrumentName
from murfey.server.api.auth import MurfeySessionIDFrontend as MurfeySessionID
from murfey.server.api.auth import validate_token
from murfey.server.api.shared import get_foil_hole as _get_foil_hole
from murfey.server.api.shared import (
    get_foil_holes_from_grid_square as _get_foil_holes_from_grid_square,
)
from murfey.server.api.shared import get_grid_squares as _get_grid_squares
from murfey.server.api.shared import (
    get_grid_squares_from_dcg as _get_grid_squares_from_dcg,
)
from murfey.server.api.shared import (
    get_machine_config_for_instrument,
    get_upstream_tiff_dirs,
    remove_session_by_id,
)
from murfey.server.ispyb import DB as ispyb_db
from murfey.server.ispyb import get_all_ongoing_visits
from murfey.server.murfey_db import murfey_db
from murfey.util import sanitise
from murfey.util.config import MachineConfig, get_machine_config
from murfey.util.db import (
    ClientEnvironment,
    DataCollection,
    DataCollectionGroup,
    FoilHole,
    GridSquare,
    Movie,
    ProcessingJob,
    RsyncInstance,
    Session,
    SPAFeedbackParameters,
    SPARelionParameters,
    Tilt,
    TiltSeries,
)
from murfey.util.models import Visit

logger = getLogger("murfey.server.api.session_info")

router = APIRouter(
    prefix="/session_info",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: General"],
)


@router.get("/health/")
def health_check(db=ispyb_db):
    conn = db.connection()
    conn.close()
    return {
        "ispyb_connection": True,
        "rabbitmq_connection": _transport_object.transport.is_connected(),
    }


@router.get("/connections/")
def connections_check():
    return {"connections": list(ws.manager.active_connections.keys())}


@router.get("/instruments/{instrument_name}/machine")
def machine_info_by_instrument(
    instrument_name: MurfeyInstrumentName,
) -> Optional[MachineConfig]:
    return get_machine_config_for_instrument(instrument_name)


@router.get("/instruments/{instrument_name}/visits_raw", response_model=List[Visit])
def get_current_visits(instrument_name: MurfeyInstrumentName, db=ispyb_db):
    logger.debug(
        f"Received request to look up ongoing visits for {sanitise(instrument_name)}"
    )
    return get_all_ongoing_visits(instrument_name, db)


@router.get("/instruments/{instrument_name}/visits/")
def all_visit_info(
    instrument_name: MurfeyInstrumentName, request: Request, db=ispyb_db
):
    visits = get_all_ongoing_visits(instrument_name, db)

    if visits:
        return_query = [
            {
                "Start date": visit.start,
                "End date": visit.end,
                "Visit name": visit.name,
                "Time remaining": str(visit.end - datetime.now()),
            }
            for visit in visits
        ]  # "Proposal title": visit.proposal_title
        logger.debug(
            f"{len(visits)} visits active for {sanitise(instrument_name)=}: {', '.join(v.name for v in visits)}"
        )
        return templates.TemplateResponse(
            request=request,
            name="activevisits.html",
            context={"info": return_query, "microscope": instrument_name},
        )
    else:
        logger.debug(f"No visits identified for {sanitise(instrument_name)=}")
        return templates.TemplateResponse(
            request=request,
            name="activevisits.html",
            context={"info": [], "microscope": instrument_name},
        )


@router.get("/sessions/{session_id}/rsyncers", response_model=List[RsyncInstance])
def get_rsyncers_for_client(session_id: MurfeySessionID, db=murfey_db):
    rsync_instances = db.exec(
        select(RsyncInstance).where(RsyncInstance.session_id == session_id)
    )
    return rsync_instances.all()


class SessionClients(BaseModel):
    session: Session
    clients: List[ClientEnvironment]


@router.get("/session/{session_id}")
async def get_session(session_id: MurfeySessionID, db=murfey_db) -> SessionClients:
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    clients = db.exec(
        select(ClientEnvironment).where(ClientEnvironment.session_id == session_id)
    ).all()
    return SessionClients(session=session, clients=clients)


@router.get("/sessions")
async def get_sessions(db=murfey_db):
    sessions = db.exec(select(Session)).all()
    clients = db.exec(select(ClientEnvironment)).all()
    res = []
    for sess in sessions:
        r = {"session": sess, "clients": []}
        for cl in clients:
            if cl.session_id == sess.id:
                r["clients"].append(cl)
        res.append(r)
    return res


class VisitEndTime(BaseModel):
    end_time: Optional[datetime] = None


@router.post("/instruments/{instrument_name}/visits/{visit}/session/{name}")
def create_session(
    instrument_name: MurfeyInstrumentName,
    visit: str,
    name: str,
    visit_end_time: VisitEndTime,
    db=murfey_db,
) -> int:
    s = Session(
        name=name,
        visit=visit,
        instrument_name=instrument_name,
        visit_end_time=visit_end_time.end_time,
    )
    db.add(s)
    db.commit()
    sid = s.id
    return sid


@router.post("/sessions/{session_id}")
def update_session(
    session_id: MurfeySessionID, process: bool = True, db=murfey_db
) -> None:
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    session.process = process
    db.add(session)
    db.commit()
    return None


@router.delete("/sessions/{session_id}")
def remove_session(session_id: MurfeySessionID, db=murfey_db):
    remove_session_by_id(session_id, db)


@router.get("/instruments/{instrument_name}/visits/{visit_name}/sessions")
def get_sessions_with_visit(
    instrument_name: MurfeyInstrumentName, visit_name: str, db=murfey_db
) -> List[Session]:
    sessions = db.exec(
        select(Session)
        .where(Session.instrument_name == instrument_name)
        .where(Session.visit == visit_name)
    ).all()
    return sessions


@router.get("/instruments/{instrument_name}/sessions")
async def get_sessions_by_instrument_name(
    instrument_name: MurfeyInstrumentName, db=murfey_db
) -> List[Session]:
    sessions = db.exec(
        select(Session).where(Session.instrument_name == instrument_name)
    ).all()
    return sessions


@router.get("/sessions/{session_id}/data_collection_groups")
def get_dc_groups(
    session_id: MurfeySessionID, db=murfey_db
) -> Dict[str, DataCollectionGroup]:
    data_collection_groups = db.exec(
        select(DataCollectionGroup).where(DataCollectionGroup.session_id == session_id)
    ).all()
    return {dcg.tag: dcg for dcg in data_collection_groups}


@router.get("/sessions/{session_id}/data_collection_groups/{dcgid}/data_collections")
def get_data_collections(
    session_id: MurfeySessionID, dcgid: int, db=murfey_db
) -> List[DataCollection]:
    data_collections = db.exec(
        select(DataCollection).where(DataCollection.dcg_id == dcgid)
    ).all()
    return data_collections


@router.get("/clients")
async def get_clients(db=murfey_db):
    clients = db.exec(select(ClientEnvironment)).all()
    return clients


class CurrentGainRef(BaseModel):
    path: str


@router.put("/sessions/{session_id}/current_gain_ref")
def update_current_gain_ref(
    session_id: MurfeySessionID, new_gain_ref: CurrentGainRef, db=murfey_db
):
    session = db.exec(select(Session).where(Session.id == session_id)).one()
    session.current_gain_ref = new_gain_ref.path
    db.add(session)
    db.commit()


spa_router = APIRouter(
    prefix="/session_info/spa",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: SPA"],
)


class ProcessingDetails(BaseModel):
    data_collection_group: DataCollectionGroup
    data_collections: List[DataCollection]
    processing_jobs: List[ProcessingJob]
    relion_params: SPARelionParameters
    feedback_params: SPAFeedbackParameters


@spa_router.get("/sessions/{session_id}/spa_processing_parameters")
def get_spa_proc_param_details(
    session_id: MurfeySessionID, db=murfey_db
) -> Optional[List[ProcessingDetails]]:
    params = db.exec(
        select(
            DataCollectionGroup,
            DataCollection,
            ProcessingJob,
            SPARelionParameters,
            SPAFeedbackParameters,
        )
        .where(DataCollectionGroup.session_id == session_id)
        .where(DataCollectionGroup.id == DataCollection.dcg_id)
        .where(DataCollection.id == ProcessingJob.dc_id)
        .where(SPARelionParameters.pj_id == ProcessingJob.id)
        .where(SPAFeedbackParameters.pj_id == ProcessingJob.id)
    ).all()
    if not params:
        return None
    unique_dcg_indices = []
    dcg_ids = []
    for i, p in enumerate(params):
        if p[0].id not in dcg_ids:
            dcg_ids.append(p[0].id)
            unique_dcg_indices.append(i)

    def _parse(ps, i, dcg_id):
        res = []
        for p in ps:
            if p[0].id == dcg_id:
                if p[i] not in res:
                    res.append(p[i])
        return res

    return [
        ProcessingDetails(
            data_collection_group=params[i][0],
            data_collections=_parse(params, 1, d),
            processing_jobs=_parse(params, 2, d),
            relion_params=_parse(params, 3, d)[0],
            feedback_params=_parse(params, 4, d)[0],
        )
        for i, d in zip(unique_dcg_indices, dcg_ids)
    ]


@spa_router.get(
    "/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares/{gsid}/foil_holes/{fhid}/num_movies"
)
def get_number_of_movies_from_foil_hole(
    session_id: int, dcgid: int, gsid: int, fhid: int, db=murfey_db
) -> int:
    movies = db.exec(
        select(Movie, FoilHole, GridSquare, DataCollectionGroup)
        .where(Movie.foil_hole_id == FoilHole.id)
        .where(FoilHole.name == fhid)
        .where(FoilHole.grid_square_id == GridSquare.id)
        .where(GridSquare.name == gsid)
        .where(GridSquare.session_id == session_id)
        .where(GridSquare.tag == DataCollectionGroup.tag)
        .where(DataCollectionGroup.id == dcgid)
    ).all()
    return len(movies)


@spa_router.get("/sessions/{session_id}/grid_squares")
def get_grid_squares(session_id: MurfeySessionID, db=murfey_db):
    return _get_grid_squares(session_id, db)


@spa_router.get("/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares")
def get_grid_squares_from_dcg(
    session_id: MurfeySessionID, dcgid: int, db=murfey_db
) -> List[GridSquare]:
    return _get_grid_squares_from_dcg(session_id, dcgid, db)


@spa_router.get(
    "/sessions/{session_id}/data_collection_groups/{dcgid}/grid_squares/{gsid}/foil_holes"
)
def get_foil_holes_from_grid_square(
    session_id: MurfeySessionID, dcgid: int, gsid: int, db=murfey_db
) -> List[FoilHole]:
    return _get_foil_holes_from_grid_square(session_id, dcgid, gsid, db)


@spa_router.get("/sessions/{session_id}/foil_hole/{fh_name}")
def get_foil_hole(
    session_id: MurfeySessionID, fh_name: int, db=murfey_db
) -> Dict[str, int]:
    return _get_foil_hole(session_id, fh_name, db)


tomo_router = APIRouter(
    prefix="/session_info/tomo",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: CryoET"],
)


@tomo_router.get("/sessions/{session_id}/tilt_series/{tilt_series_tag}/tilts")
def get_tilts(
    session_id: MurfeySessionID, tilt_series_tag: str, db=murfey_db
) -> Dict[str, List[str]]:
    res = db.exec(
        select(TiltSeries, Tilt)
        .where(TiltSeries.tag == tilt_series_tag)
        .where(TiltSeries.session_id == session_id)
        .where(Tilt.tilt_series_id == TiltSeries.id)
    ).all()
    tilts: Dict[str, List[str]] = {}
    for el in res:
        if tilts.get(el[1].rsync_source):
            tilts[el[1].rsync_source].append(el[2].movie_path)
        else:
            tilts[el[1].rsync_source] = [el[2].movie_path]
    return tilts


correlative_router = APIRouter(
    prefix="/session_info/correlative",
    dependencies=[Depends(validate_token)],
    tags=["Session Info: Correlative Imaging"],
)


@correlative_router.get("/sessions/{session_id}/upstream_visits")
async def find_upstream_visits(session_id: MurfeySessionID, db=murfey_db):
    murfey_session = db.exec(select(Session).where(Session.id == session_id)).one()
    visit_name = murfey_session.visit
    instrument_name = murfey_session.instrument_name
    machine_config = get_machine_config(instrument_name=instrument_name)[
        instrument_name
    ]
    upstream_visits = {}
    # Iterates through provided upstream directories
    for p in machine_config.upstream_data_directories:
        # Looks for visit name in file path
        for v in Path(p).glob(f"{visit_name.split('-')[0]}-*"):
            upstream_visits[v.name] = v / machine_config.processed_directory_name
    return upstream_visits


@correlative_router.get("/visits/{visit_name}/{session_id}/upstream_tiff_paths")
async def gather_upstream_tiffs(visit_name: str, session_id: int, db=murfey_db):
    """
    Looks for TIFF files associated with the current session in the permitted storage
    servers, and returns their relative file paths as a list.
    """
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    upstream_tiff_paths = []
    tiff_dirs = get_upstream_tiff_dirs(visit_name, instrument_name)
    if not tiff_dirs:
        return None
    for tiff_dir in tiff_dirs:
        for f in tiff_dir.glob("**/*.tiff"):
            upstream_tiff_paths.append(str(f.relative_to(tiff_dir)))
        for f in tiff_dir.glob("**/*.tif"):
            upstream_tiff_paths.append(str(f.relative_to(tiff_dir)))
    return upstream_tiff_paths


@correlative_router.get(
    "/visits/{visit_name}/{session_id}/upstream_tiff/{tiff_path:path}"
)
async def get_tiff(visit_name: str, session_id: int, tiff_path: str, db=murfey_db):
    instrument_name = (
        db.exec(select(Session).where(Session.id == session_id)).one().instrument_name
    )
    tiff_dirs = get_upstream_tiff_dirs(visit_name, instrument_name)
    if not tiff_dirs:
        return None

    tiff_path = "/".join(secure_filename(p) for p in tiff_path.split("/"))
    for tiff_dir in tiff_dirs:
        test_path = tiff_dir / tiff_path
        if test_path.is_file():
            break
    else:
        logger.warning(f"TIFF {tiff_path} not found")
        return None

    return FileResponse(path=test_path)



import json
import logging
import subprocess
import threading
import time
from dataclasses import dataclass, field
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import quote, urlparse

import requests

import murfey.client.websocket
from murfey.client.analyser import Analyser
from murfey.client.contexts.spa import SPAModularContext
from murfey.client.contexts.tomo import TomographyContext
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.client.rsync import RSyncer, RSyncerUpdate, TransferResult
from murfey.client.tui.screens import determine_default_destination
from murfey.client.watchdir import DirWatcher
from murfey.util import posix_path
from murfey.util.api import url_path_for
from murfey.util.client import capture_delete, capture_post, get_machine_config_client

log = logging.getLogger("murfey.client.multigrid_control")


@dataclass
class MultigridController:
    sources: list[Path]
    visit: str
    instrument_name: str
    session_id: int
    murfey_url: str = "http://localhost:8000"
    rsync_url: str = ""
    rsync_module: str = "data"
    demo: bool = False
    finalising: bool = False
    dormant: bool = False
    multigrid_watcher_active: bool = True
    processing_enabled: bool = True
    do_transfer: bool = True
    dummy_dc: bool = False
    force_mdoc_metadata: bool = True
    rsync_restarts: List[str] = field(default_factory=lambda: [])
    rsync_processes: Dict[Path, RSyncer] = field(default_factory=lambda: {})
    analysers: Dict[Path, Analyser] = field(default_factory=lambda: {})
    data_collection_parameters: dict = field(default_factory=lambda: {})
    token: str = ""
    _machine_config: dict = field(default_factory=lambda: {})
    visit_end_time: Optional[datetime] = None

    def __post_init__(self):
        if self.token:
            requests.get = partial(
                requests.get, headers={"Authorization": f"Bearer {self.token}"}
            )
            requests.post = partial(
                requests.post, headers={"Authorization": f"Bearer {self.token}"}
            )
            requests.delete = partial(
                requests.delete, headers={"Authorization": f"Bearer {self.token}"}
            )
        machine_data = requests.get(
            f"{self.murfey_url}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=self.instrument_name)}"
        ).json()
        self.rsync_url = machine_data.get("rsync_url", "")
        self.rsync_module = machine_data.get("rsync_module", "data")
        self._environment = MurfeyInstanceEnvironment(
            url=urlparse(self.murfey_url, allow_fragments=False),
            client_id=0,
            murfey_session=self.session_id,
            software_versions=machine_data.get("software_versions", {}),
            demo=self.demo,
            visit=self.visit,
            dose_per_frame=self.data_collection_parameters.get("dose_per_frame"),
            gain_ref=self.data_collection_parameters.get("gain_ref"),
            symmetry=self.data_collection_parameters.get("symmetry"),
            eer_fractionation=self.data_collection_parameters.get("eer_fractionation"),
            instrument_name=self.instrument_name,
            # processing_only_mode=server_routing_prefix_found,
        )
        self._machine_config = get_machine_config_client(
            str(self._environment.url.geturl()),
            instrument_name=self._environment.instrument_name,
            demo=self._environment.demo,
        )
        self._data_suffixes = (".mrc", ".tiff", ".tif", ".eer")
        self._data_substrings = [
            s
            for val in self._machine_config["data_required_substrings"].values()
            for ds in val.values()
            for s in ds
        ]
        self._data_collection_form_complete = False
        self._register_dc: bool | None = None
        self.rsync_processes = self.rsync_processes or {}
        self.analysers = self.analysers or {}

        self.ws = murfey.client.websocket.WSApp(
            server=self.murfey_url,
            register_client=False,
        )

        # Calculate the time offset between the client and the server
        current_time = datetime.now()
        server_timestamp = requests.get(
            f"{self.murfey_url}{url_path_for('session_control.router', 'get_current_timestamp')}"
        ).json()["timestamp"]
        self.server_time_offset = current_time - datetime.fromtimestamp(
            server_timestamp
        )

        # Store the visit end time in the current device's equivalent time
        if self.visit_end_time:
            self.visit_end_time += self.server_time_offset

    def _multigrid_watcher_finalised(self):
        self.multigrid_watcher_active = False

    def is_ready_for_dormancy(self):
        """
        When the multigrid watcher is no longer active, sends a request to safely stop
        the analyser and file watcher threads, then checks to see that those threads
        and the RSyncer processes associated with the current session have all been
        safely stopped
        """
        log.debug(
            f"Starting dormancy check for MultigridController for session {self.session_id}"
        )
        if not self.multigrid_watcher_active:
            for a in self.analysers.values():
                if a.is_safe_to_stop():
                    a.stop()
            for w in self._environment.watchers.values():
                if w.is_safe_to_stop():
                    w.stop()
            return (
                all(r._finalised for r in self.rsync_processes.values())
                and not any(a.thread.is_alive() for a in self.analysers.values())
                and not any(
                    w.thread.is_alive() for w in self._environment.watchers.values()
                )
            )
        log.debug(f"Multigrid watcher for session {self.session_id} is still active")
        return False

    def clean_up_once_dormant(self, running_threads: list[threading.Thread]):
        """
        A function run in a separate thread that runs the post-session cleanup logic
        once all threads associated with this current session are halted, and marks
        the controller as being fully dormant after doing so.
        """
        for thread in running_threads:
            thread.join()
            log.debug(f"RSyncer cleanup thread {thread.ident} has stopped safely")
        while not self.is_ready_for_dormancy():
            time.sleep(10)

        # Once all threads are stopped, remove session from the database
        log.debug(
            f"Submitting request to remove session {self.session_id} from database"
        )
        response = capture_delete(
            f"{self._environment.url.geturl()}{url_path_for('session_control.router', 'remove_session', session_id=self.session_id)}",
        )
        success = response.status_code == 200 if response else False
        if not success:
            log.warning(f"Could not delete database data for {self.session_id}")

        # Send message to frontend to trigger a refresh
        self.ws.send(
            json.dumps(
                {
                    "message": "refresh",
                    "target": "sessions",
                    "instrument_name": self.instrument_name,
                }
            )
        )

        # Mark as dormant
        self.dormant = True

    def abandon(self):
        for a in self.analysers.values():
            a.request_stop()
        for w in self._environment.watchers.values():
            w.request_stop()
        for p in self.rsync_processes.values():
            p.request_stop()

    def finalise(self):
        self.finalising = True
        for a in self.analysers.values():
            a.request_stop()
            log.debug(f"Stop request sent to analyser {a}")
        for w in self._environment.watchers.values():
            w.request_stop()
            log.debug(f"Stop request sent to watcher {w}")
        rsync_finaliser_threads = []
        for p in self.rsync_processes.keys():
            # Collect the running rsyncer finaliser threads to pass to the dormancy checker
            rsync_finaliser_threads.append(self._finalise_rsyncer(p))
            log.debug(f"Finalised rsyncer {p}")

        # Run the session cleanup function in a separate thread
        cleanup_upon_dormancy_thread = threading.Thread(
            target=self.clean_up_once_dormant,
            args=[rsync_finaliser_threads],
            daemon=True,
        )
        cleanup_upon_dormancy_thread.start()

    def update_visit_time(self, new_end_time: datetime):
        # Convert the received server timestamp into the local equivalent
        self.visit_end_time = new_end_time + self.server_time_offset
        for rp in self.rsync_processes.values():
            rp._end_time = self.visit_end_time

    def _start_rsyncer_multigrid(
        self,
        source: Path,
        extra_directory: str = "",
        include_mid_path: bool = True,
        use_suggested_path: bool = True,
        destination_overrides: Optional[Dict[Path, str]] = None,
        remove_files: bool = False,
        analyse: bool = True,
        tag: str = "",
        limited: bool = False,
    ):
        log.info(f"Starting multigrid rsyncer: {source}")
        log.debug(f"Analysis of {source} is {('enabled' if analyse else 'disabled')}")
        destination_overrides = destination_overrides or {}
        machine_data = requests.get(
            f"{self._environment.url.geturl()}{url_path_for('session_control.router', 'machine_info_by_instrument', instrument_name=self.instrument_name)}"
        ).json()
        if destination_overrides.get(source):
            destination = (
                destination_overrides[source]
                if str(source) in self.rsync_restarts
                else destination_overrides[source] + f"/{extra_directory}"
            )
        else:
            for k, v in destination_overrides.items():
                if Path(v).name in source.parts:
                    destination = str(k / extra_directory)
                    break
            else:
                self._environment.default_destinations[source] = (
                    f"{datetime.now().year}"
                )
                destination = determine_default_destination(
                    self._environment.visit,
                    source,
                    self._environment.default_destinations[source],
                    self._environment,
                    self.analysers or {},
                    touch=True,
                    extra_directory=extra_directory,
                    include_mid_path=include_mid_path,
                    use_suggested_path=use_suggested_path,
                )
        self._environment.sources.append(source)
        self._start_rsyncer(
            source,
            destination,
            force_metadata=self.processing_enabled,
            analyse=analyse,
            remove_files=remove_files,
            tag=tag,
            limited=limited,
            transfer=machine_data.get("data_transfer_enabled", True),
            restarted=str(source) in self.rsync_restarts,
        )
        self.ws.send(
            json.dumps(
                {
                    "message": "refresh",
                    "target": "rsyncer",
                    "session_id": self.session_id,
                }
            )
        )

    def _rsyncer_stopped(self, source: Path, explicit_stop: bool = False):
        if explicit_stop:
            remove_url = f"{self.murfey_url}{url_path_for('session_control.router', 'delete_rsyncer', session_id=self.session_id)}?source={quote(str(source), safe='')}"
            requests.delete(remove_url)
        else:
            stop_url = f"{self.murfey_url}{url_path_for('session_control.router', 'register_stopped_rsyncer', session_id=self.session_id)}"
            capture_post(stop_url, json={"source": str(source)})

    def _finalise_rsyncer(self, source: Path):
        """
        Starts a new Rsyncer thread that cleans up the directories, and returns that
        thread to be managed by a central thread.
        """
        finalise_thread = threading.Thread(
            name=f"Controller finaliser thread ({source})",
            target=self.rsync_processes[source].finalise,
            kwargs={"thread": False},
            daemon=True,
        )
        finalise_thread.start()
        log.debug(f"Started RSync cleanup for {str(source)}")
        return finalise_thread

    def _restart_rsyncer(self, source: Path):
        self.rsync_processes[source].restart()
        restarted_url = f"{self.murfey_url}{url_path_for('session_control.router', 'register_restarted_rsyncer', session_id=self.session_id)}"
        capture_post(restarted_url, json={"source": str(source)})

    def _request_watcher_stop(self, source: Path):
        self._environment.watchers[source]._stopping = True
        self._environment.watchers[source]._halt_thread = True

    def _start_rsyncer(
        self,
        source: Path,
        destination: str,
        visit_path: str = "",
        force_metadata: bool = False,
        analyse: bool = True,
        remove_files: bool = False,
        tag: str = "",
        limited: bool = False,
        transfer: bool = True,
        restarted: bool = False,
    ):
        log.info(f"starting rsyncer: {source}")
        if transfer:
            # Always make sure the destination directory exists
            make_directory_url = f"{self.murfey_url}{url_path_for('file_io_instrument.router', 'make_rsyncer_destination', session_id=self.session_id)}"
            capture_post(make_directory_url, json={"destination": destination})
        if self._environment:
            self._environment.default_destinations[source] = destination
            if self._environment.gain_ref and visit_path:
                # Set up rsync command
                rsync_cmd = [
                    "rsync",
                    f"{posix_path(self._environment.gain_ref)!r}",  # '!r' will print strings in ''
                    f"{self._environment.url.hostname}::{self.rsync_module}/{visit_path}/processing",
                ]
                # Wrap in bash shell
                cmd = [
                    "bash",
                    "-c",
                    " ".join(rsync_cmd),
                ]
                # Run rsync subprocess
                gain_rsync = subprocess.run(cmd)
                if gain_rsync.returncode:
                    log.warning(
                        f"Gain reference file {posix_path(self._environment.gain_ref)!r} was not successfully transferred to {visit_path}/processing"
                    )
        if transfer:
            self.rsync_processes[source] = RSyncer(
                source,
                basepath_remote=Path(destination),
                rsync_module=self.rsync_module,
                server_url=(
                    urlparse(self.rsync_url)
                    if self.rsync_url
                    else self._environment.url
                ),
                stop_callback=self._rsyncer_stopped,
                do_transfer=self.do_transfer,
                remove_files=remove_files,
                end_time=self.visit_end_time,
            )

            def rsync_result(update: RSyncerUpdate):
                if not update.base_path:
                    raise ValueError("No base path from rsyncer update")
                if not self.rsync_processes.get(update.base_path):
                    raise ValueError("TUI rsync process does not exist")
                if update.outcome is TransferResult.SUCCESS:
                    # log.info(
                    #     f"File {str(update.file_path)!r} successfully transferred ({update.file_size} bytes)"
                    # )
                    pass
                else:
                    log.warning(f"Failed to transfer file {str(update.file_path)!r}")
                    self.rsync_processes[update.base_path].enqueue(update.file_path)

            self.rsync_processes[source].subscribe(rsync_result)
            self.rsync_processes[source].subscribe(
                partial(
                    self._increment_transferred_files_prometheus,
                    destination=destination,
                    source=str(source),
                )
            )
            self.rsync_processes[source].subscribe(
                partial(
                    self._increment_transferred_files,
                    destination=destination,
                    source=str(source),
                ),
                secondary=True,
            )
            if restarted:
                restarted_url = f"{self.murfey_url}{url_path_for('session_control.router', 'register_restarted_rsyncer', session_id=self.session_id)}"
                capture_post(restarted_url, json={"source": str(source)})
            else:
                url = f"{str(self._environment.url.geturl())}{url_path_for('session_control.router', 'register_rsyncer', session_id=self._environment.murfey_session)}"
                rsyncer_data = {
                    "source": str(source),
                    "destination": destination,
                    "session_id": self.session_id,
                    "transferring": self.do_transfer or self._environment.demo,
                    "tag": tag,
                }
                requests.post(url, json=rsyncer_data)
        self._environment.watchers[source] = DirWatcher(source, settling_time=30)

        if not self.analysers.get(source) and analyse:
            log.info(f"Starting analyser for {source}")
            self.analysers[source] = Analyser(
                source,
                environment=self._environment if not self.dummy_dc else None,
                force_mdoc_metadata=self.force_mdoc_metadata,
                limited=limited,
            )
            if force_metadata:
                self.analysers[source].subscribe(
                    partial(self._start_dc, from_form=True)
                )
            else:
                self.analysers[source].subscribe(self._data_collection_form)
            self.analysers[source].start()
            if transfer:
                self.rsync_processes[source].subscribe(self.analysers[source].enqueue)

        if transfer:
            self.rsync_processes[source].start()

        if self._environment:
            if self._environment.watchers.get(source):
                if transfer:
                    self._environment.watchers[source].subscribe(
                        self.rsync_processes[source].enqueue
                    )
                else:
                    # the watcher and rsyncer don't notify with the same object so conversion required here
                    def _rsync_update_converter(p: Path) -> None:
                        self.analysers[source].enqueue(
                            RSyncerUpdate(
                                file_path=p,
                                file_size=0,
                                outcome=TransferResult.SUCCESS,
                                transfer_total=0,
                                queue_size=0,
                                base_path=source,
                            )
                        )
                        return None

                    self._environment.watchers[source].subscribe(
                        _rsync_update_converter
                    )
                self._environment.watchers[source].subscribe(
                    partial(
                        self._increment_file_count,
                        destination=destination,
                        source=str(source),
                    ),
                    secondary=True,
                )
                self._environment.watchers[source].start()

    def _data_collection_form(self, response: dict):
        log.info("data collection form ready")
        if self._data_collection_form_complete:
            return
        if self._register_dc and response.get("form"):
            self._form_values = {k: str(v) for k, v in response.get("form", {}).items()}
            log.info(
                f"gain reference is set to {self._form_values.get('gain_ref')}, {self._environment.gain_ref}"
            )
            if self._form_values.get("gain_ref") in (None, "None"):
                self._form_values["gain_ref"] = self._environment.gain_ref
            self._data_collection_form_complete = True
        elif self._register_dc is None:
            self._data_collection_form_complete = True

    def _start_dc(self, metadata_json, from_form: bool = False):
        if self.dummy_dc:
            return
        # for multigrid the analyser sends the message straight to _start_dc by-passing user input
        # it is then necessary to extract the data from the message
        if from_form:
            metadata_json = metadata_json.get("form", {})
            # Safely convert all entries into strings, but leave None as-is
            metadata_json = {
                k: str(v) if v is not None else None for k, v in metadata_json.items()
            }
        self._environment.dose_per_frame = metadata_json.get("dose_per_frame")
        self._environment.gain_ref = metadata_json.get("gain_ref")
        self._environment.symmetry = metadata_json.get("symmetry")
        self._environment.eer_fractionation = metadata_json.get("eer_fractionation")
        source = Path(metadata_json["source"])
        context = self.analysers[source]._context
        if context:
            context.data_collection_parameters = {
                k: None if v == "None" else v for k, v in metadata_json.items()
            }
        if isinstance(context, TomographyContext):
            source = Path(metadata_json["source"])
            context.register_tomography_data_collections(
                file_extension=metadata_json["file_extension"],
                image_directory=str(self._environment.default_destinations[source]),
                environment=self._environment,
            )

            log.info("Registering tomography processing parameters")
            if context.data_collection_parameters.get("num_eer_frames"):
                eer_response = requests.post(
                    f"{str(self._environment.url.geturl())}{url_path_for('file_io_instrument.router', 'write_eer_fractionation_file', visit_name=self._environment.visit, session_id=self._environment.murfey_session)}",
                    json={
                        "num_frames": context.data_collection_parameters[
                            "num_eer_frames"
                        ],
                        "fractionation": self._environment.eer_fractionation,
                        "dose_per_frame": self._environment.dose_per_frame,
                        "fractionation_file_name": "eer_fractionation_tomo.txt",
                    },
                )
                eer_fractionation_file = eer_response.json()["eer_fractionation_file"]
                metadata_json.update({"eer_fractionation_file": eer_fractionation_file})
            capture_post(
                f"{self._environment.url.geturl()}{url_path_for('workflow.tomo_router', 'register_tomo_proc_params', session_id=self._environment.murfey_session)}",
                json=metadata_json,
            )
            capture_post(
                f"{self._environment.url.geturl()}{url_path_for('workflow.tomo_router', 'flush_tomography_processing', visit_name=self._environment.visit, session_id=self._environment.murfey_session)}",
                json={"rsync_source": str(source)},
            )
            log.info("Tomography processing flushed")

        elif isinstance(context, SPAModularContext):
            url = f"{str(self._environment.url.geturl())}{url_path_for('workflow.router', 'register_dc_group', visit_name=self._environment.visit, session_id=self.session_id)}"
            dcg_data = {
                "experiment_type": "single particle",
                "experiment_type_id": 37,
                "tag": str(source),
                "atlas": (
                    str(self._environment.samples[source].atlas)
                    if self._environment.samples.get(source)
                    else ""
                ),
                "sample": (
                    self._environment.samples[source].sample
                    if self._environment.samples.get(source)
                    else None
                ),
            }
            capture_post(url, json=dcg_data)
            if from_form:
                data = {
                    "voltage": metadata_json["voltage"],
                    "pixel_size_on_image": metadata_json["pixel_size_on_image"],
                    "experiment_type": metadata_json["experiment_type"],
                    "image_size_x": metadata_json["image_size_x"],
                    "image_size_y": metadata_json["image_size_y"],
                    "file_extension": metadata_json["file_extension"],
                    "acquisition_software": metadata_json["acquisition_software"],
                    "image_directory": str(
                        self._environment.default_destinations[source]
                    ),
                    "tag": str(source),
                    "source": str(source),
                    "magnification": metadata_json["magnification"],
                    "total_exposed_dose": metadata_json.get("total_exposed_dose"),
                    "c2aperture": metadata_json.get("c2aperture"),
                    "exposure_time": metadata_json.get("exposure_time"),
                    "slit_width": metadata_json.get("slit_width"),
                    "phase_plate": metadata_json.get("phase_plate", False),
                }
                capture_post(
                    f"{str(self._environment.url.geturl())}{url_path_for('workflow.router', 'start_dc', visit_name=self._environment.visit, session_id=self.session_id)}",
                    json=data,
                )
                for recipe in (
                    "em-spa-preprocess",
                    "em-spa-extract",
                    "em-spa-class2d",
                    "em-spa-class3d",
                    "em-spa-refine",
                ):
                    capture_post(
                        f"{str(self._environment.url.geturl())}{url_path_for('workflow.router', 'register_proc', visit_name=self._environment.visit, session_id=self.session_id)}",
                        json={
                            "tag": str(source),
                            "source": str(source),
                            "recipe": recipe,
                        },
                    )
                log.info(f"Posting SPA processing parameters: {metadata_json}")
                response = capture_post(
                    f"{self._environment.url.geturl()}{url_path_for('workflow.spa_router', 'register_spa_proc_params', session_id=self.session_id)}",
                    json={
                        **{
                            k: None if v == "None" else v
                            for k, v in metadata_json.items()
                        },
                        "tag": str(source),
                    },
                )
                if response and not str(response.status_code).startswith("2"):
                    log.warning(f"{response.reason}")
                capture_post(
                    f"{self._environment.url.geturl()}{url_path_for('workflow.spa_router', 'flush_spa_processing', visit_name=self._environment.visit, session_id=self.session_id)}",
                    json={"tag": str(source)},
                )

    def _increment_file_count(
        self, observed_files: List[Path], source: str, destination: str
    ):
        url = f"{str(self._environment.url.geturl())}{url_path_for('prometheus.router', 'increment_rsync_file_count', visit_name=self._environment.visit)}"
        num_data_files = len(
            [
                f
                for f in observed_files
                if f.suffix in self._data_suffixes
                and any(substring in f.name for substring in self._data_substrings)
            ]
        )
        data = {
            "source": source,
            "destination": destination,
            "session_id": self.session_id,
            "increment_count": len(observed_files),
            "increment_data_count": num_data_files,
        }
        requests.post(url, json=data)

    # Prometheus can handle higher traffic so update for every transferred file rather
    # than batching as we do for the Murfey database updates in _increment_transferred_files
    def _increment_transferred_files_prometheus(
        self, update: RSyncerUpdate, source: str, destination: str
    ):
        if update.outcome is TransferResult.SUCCESS:
            url = f"{str(self._environment.url.geturl())}{url_path_for('prometheus.router', 'increment_rsync_transferred_files_prometheus', visit_name=self._environment.visit)}"
            data_files = (
                [update]
                if update.file_path.suffix in self._data_suffixes
                and any(
                    substring in update.file_path.name
                    for substring in self._data_substrings
                )
                else []
            )
            data = {
                "source": source,
                "destination": destination,
                "session_id": self.session_id,
                "increment_count": 1,
                "bytes": update.file_size,
                "increment_data_count": len(data_files),
                "data_bytes": sum(f.file_size for f in data_files),
            }
            requests.post(url, json=data)

    def _increment_transferred_files(
        self, updates: List[RSyncerUpdate], source: str, destination: str
    ):
        checked_updates = [
            update for update in updates if update.outcome is TransferResult.SUCCESS
        ]
        if not checked_updates:
            return
        url = f"{str(self._environment.url.geturl())}{url_path_for('prometheus.router', 'increment_rsync_transferred_files', visit_name=self._environment.visit)}"
        data_files = [
            u
            for u in updates
            if u.file_path.suffix in self._data_suffixes
            and any(
                substring in u.file_path.name for substring in self._data_substrings
            )
        ]
        data = {
            "source": source,
            "destination": destination,
            "session_id": self.session_id,
            "increment_count": len(checked_updates),
            "bytes": sum(f.file_size for f in checked_updates),
            "increment_data_count": len(data_files),
            "data_bytes": sum(f.file_size for f in data_files),
        }
        requests.post(url, json=data)



from typing import Literal
from unittest import mock
from unittest.mock import AsyncMock, MagicMock

from fastapi import FastAPI
from fastapi.testclient import TestClient
from pytest_mock import MockerFixture

from murfey.server.api.auth import validate_frontend_session_access, validate_token
from murfey.server.api.instrument import router as backend_router
from murfey.server.murfey_db import murfey_db_session
from murfey.util.api import url_path_for


def mock_aiohttp_clientsession(
    mocker: MockerFixture,
    method: Literal["get", "post", "delete"] = "get",
    json_data={},
    status=200,
):
    """
    Helper function to patch a aiohttp.ClientSession GET request. This returns a
    mocked async context manager with a mocked response that, in turn, returns
    the given JSON data and status.

    Returns the mocked ClientSession, which can then be inspected to assert that
    the expected calls were made.
    """

    # Mock out the async response
    mock_response = MagicMock()
    mock_response.json = AsyncMock(return_value=json_data)
    mock_response.status = status

    # Mock out the context manager returned by clientsession.get()
    mock_context_manager = MagicMock()
    mock_context_manager.__aenter__ = AsyncMock(return_value=mock_response)
    mock_context_manager.__aexit__ = AsyncMock(return_value=None)

    # Mock the client session
    mock_clientsession = MagicMock()
    mock_clientsession.__aenter__ = AsyncMock(return_value=mock_clientsession)
    mock_clientsession.__aexit__ = AsyncMock(return_value=None)

    # Assign the context manager to the request method being tested
    getattr(mock_clientsession, method.lower()).return_value = mock_context_manager

    # Patch 'aiohttp.ClientSession' to return the mocked client session
    mocker.patch("aiohttp.ClientSession", return_value=mock_clientsession)

    return mock_clientsession, mock_response


def test_check_multigrid_controller_status(mocker: MockerFixture):
    # Set up the objects to mock
    instrument_name = "test"
    session_id = 1
    instrment_server_url = "https://murfey.instrument-server.test"

    # Override the database session generator
    mock_session = MagicMock()
    mock_session.instrument_name = instrument_name
    mock_query_result = MagicMock()
    mock_query_result.one.return_value = mock_session
    mock_db_session = MagicMock()
    mock_db_session.exec.return_value = mock_query_result

    def mock_get_db_session():
        yield mock_db_session

    # Mock the machine config
    mock_machine_config = MagicMock()
    mock_machine_config.instrument_server_url = instrment_server_url
    mock_get_machine_config = mocker.patch(
        "murfey.server.api.instrument.get_machine_config"
    )
    mock_get_machine_config.return_value = {
        instrument_name: mock_machine_config,
    }

    # Mock the instrument server tokens dictionary
    mock_tokens = mocker.patch(
        "murfey.server.api.instrument.instrument_server_tokens",
        {session_id: {"access_token": mock.sentinel}},
    )

    # Mock out the async GET request in the endpoint
    mock_clientsession, _ = mock_aiohttp_clientsession(
        mocker,
        method="get",
        json_data={"exists": True},
        status=200,
    )

    # Set up the backend server
    backend_app = FastAPI()

    # Override validation and database dependencies
    backend_app.dependency_overrides[validate_token] = lambda: None
    backend_app.dependency_overrides[validate_frontend_session_access] = (
        lambda: session_id
    )
    backend_app.dependency_overrides[murfey_db_session] = mock_get_db_session
    backend_app.include_router(backend_router)
    backend_server = TestClient(backend_app)

    # Construct the URL paths for poking and sending to
    backend_url_path = url_path_for(
        "api.instrument.router",
        "check_multigrid_controller_status",
        session_id=session_id,
    )
    client_url_path = url_path_for(
        "api.router",
        "check_multigrid_controller_status",
        session_id=session_id,
    )

    # Poke the backend
    response = backend_server.get(backend_url_path)

    # Check that the expected calls were made
    mock_db_session.exec.assert_called_once()
    mock_get_machine_config.assert_called_once_with(instrument_name=instrument_name)
    mock_clientsession.get.assert_called_once_with(
        f"{instrment_server_url}{client_url_path}",
        headers={"Authorization": f"Bearer {mock_tokens[session_id]['access_token']}"},
    )
    assert response.status_code == 200
    assert response.json() == {"exists": True}



import json
import os
from pathlib import Path
from unittest import mock

import pytest

from murfey.util.client import read_config, set_default_acquisition_output

test_read_config_params_matrix = (
    # Environment variable to set | Append to tmp_path
    (
        "MURFEY_CLIENT_CONFIGURATION",
        "config/murfey-client-config.cfg",
    ),
    (
        "MURFEY_CLIENT_CONFIG_HOME",
        "config",
    ),
    (
        "",
        "",
    ),  # Test default home directory
)


@pytest.mark.parametrize("test_params", test_read_config_params_matrix)
def test_read_config(
    test_params: tuple[str, str],
    tmp_path,
    mock_client_configuration,
):
    # Unpack test params
    env_var, partial_path = test_params

    # Construct the environment variable and the expected config file path
    env_var_dict: dict[str, str] = {}
    if env_var:
        full_path = tmp_path / partial_path
        env_var_dict[env_var] = str(full_path)
        file_path = full_path if full_path.suffix else full_path / ".murfey"
    else:
        file_path = Path().home() / ".murfey"

    # Make directories all the way to the requested place
    file_path.parent.mkdir(parents=True, exist_ok=True)

    # Write the client config fixture to the specified file
    with open(file_path, "w") as file:
        mock_client_configuration.write(file)

    # Patch the OS environment variable and run the function
    with mock.patch.dict(os.environ, env_var_dict, clear=False):
        config = read_config()

    # Compare returned config with mock one
    assert dict(config["Murfey"]) == dict(mock_client_configuration["Murfey"])


def test_set_default_acquisition_output_normal_operation(tmp_path):
    output_dir = tmp_path / "settings.json"
    settings_json = {
        "a": {
            "b": {"data_dir": str(tmp_path)},
            "c": {
                "d": 1,
            },
        }
    }
    with open(output_dir, "w") as sf:
        json.dump(settings_json, sf)
    set_default_acquisition_output(
        tmp_path / "visit", {str(tmp_path / "settings.json"): ["a", "b", "data_dir"]}
    )
    assert (tmp_path / "_murfey_settings.json").is_file()
    with open(output_dir, "r") as sf:
        data = json.load(sf)
    assert data["a"]["b"]["data_dir"] == str(tmp_path / "visit")
    assert data["a"]["c"]["d"] == 1
    with open(output_dir.parent / "_murfey_settings.json", "r") as sf:
        data = json.load(sf)
    assert data["a"]["b"]["data_dir"] == str(tmp_path)
    assert data["a"]["c"]["d"] == 1


def test_set_default_acquisition_output_no_file_copy(tmp_path):
    output_dir = tmp_path / "settings.json"
    settings_json = {
        "a": {
            "b": {"data_dir": str(tmp_path)},
            "c": {
                "d": 1,
            },
        }
    }
    with open(output_dir, "w") as sf:
        json.dump(settings_json, sf)
    set_default_acquisition_output(
        tmp_path / "visit",
        {str(tmp_path / "settings.json"): ["a", "b", "data_dir"]},
        safe=False,
    )
    assert not (tmp_path / "_murfey_settings.json").is_file()
    with open(output_dir, "r") as sf:
        data = json.load(sf)
    assert data["a"]["b"]["data_dir"] == str(tmp_path / "visit")
    assert data["a"]["c"]["d"] == 1
